
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absgra.h"


abstractGraph::abstractGraph() throw()
{
    LogEntry(LOG_MEM,"...Abstract graph object allocated");
}


abstractGraph::~abstractGraph() throw()
{
    LogEntry(LOG_MEM,"...Abstract graph object disallocated");
}


unsigned long abstractGraph::Allocated() throw()
{
    return 0;
}                           


void abstractGraph::WriteGraph(const char* wrFile,TWriteInc options)
    throw(ERFile)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    sprintf(CT.logBuffer,"Writing graph to \"%s\"...",wrFile);
    LogEntry(LOG_IO,CT.logBuffer);
    
    goblinExport F(wrFile,CT);
    
    if (options==WRITE_INCIDENCES) F.StartTuple("graph",0);
    else  F.StartTuple("dense_graph",0);
    
    F.StartTuple("definition",0);
    
    F.StartTuple("nodes",1);
    F.MakeIntItem(n,0);
    F.MakeIntItem(n,0);
    F.MakeIntItem(ni,0);
    F.EndTuple();
    
    if (options==WRITE_INCIDENCES)
    {
        F.StartTuple("arcs",1);
        F.MakeIntItem(m,0);
        F.EndTuple();
    
        WriteIncidences(&F);
    }

    WriteUCap(&F);
    WriteLCap(&F);
    WriteDemand(&F);
    WriteOrientation(&F);
    
    F.EndTuple();   // definition
    
    WriteObjectives(&F);
    WriteGeometry(&F);
    WriteLayout(&F);
    WriteSolutions(&F);
    
    CT.sourceNode = Source();
    CT.targetNode = Target();
    CT.rootNode = Root();
    F.WriteConfiguration(this);

    F.EndTuple();   // graph

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}
