
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, May 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absobj.h"


goblinDataObject::goblinDataObject(goblinController &thisContext,
    TOption options) throw() : CT(thisContext)
{
    refCounter = 0;

    if (!options) OH = CT.InsertObject(this);
    else OH = NoHandle;
    
    label = NULL;
    SetLabel("unnamed");

    LogEntry(LOG_MEM,"...Abstract data object inserted into context");
}


goblinDataObject::~goblinDataObject() throw()
{
    #if defined(_FAILSAVE_)
             
    if (refCounter)
    {
        InternalError("goblinDataObject","Object is referenced");
    }
             
    #endif   

    if (OH!=NoHandle) CT.DeleteObject(OH);
    
    SetLabel(NULL);

    LogEntry(LOG_MEM,"...Abstract data object disallocated");
}


unsigned long goblinDataObject::Allocated() throw()
{
    return strlen(label)+1;
}


char* goblinDataObject::Display() throw(ERFile,ERRejected)
{
    Error(ERR_REJECTED,"Display","This object cannot be viewed");

    throw ERRejected();
}


void goblinDataObject::NoSuchNode(char* methodName,TNode v) throw(ERRange)
{
    if (v==NoNode) sprintf(CT.logBuffer,"Undefined node");
    else sprintf(CT.logBuffer,"No such node: %ld",v);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoMoreArcs(char* methodName,TNode v) throw(ERRejected)
{
    sprintf(CT.logBuffer,"No more arcs: %ld",v);

    CT.Error(ERR_REJECTED,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchArc(char* methodName,TArc a) throw(ERRange)
{
    if (a==NoArc) sprintf(CT.logBuffer,"Undefined arc");
    else sprintf(CT.logBuffer,"No such arc: %ld",a);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::CancelledArc(char* methodName,TArc a) throw(ERRejected)
{
    sprintf(CT.logBuffer,"Cancelled arc: %ld",a);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchHandle(char* methodName,THandle H) throw(ERRejected)
{
    if (H==NoHandle) sprintf(CT.logBuffer,"Undefined handle");
    else sprintf(CT.logBuffer,"No such handle: %ld",H);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchItem(char* methodName,unsigned long i) throw(ERRange)
{
    sprintf(CT.logBuffer,"No such item: %ld",i);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchItem(char* methodName,unsigned short i) throw(ERRange)
{
    sprintf(CT.logBuffer,"No such item: %d",i);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchIndex(char* methodName,unsigned long i) throw(ERRange)
{
    sprintf(CT.logBuffer,"No such index: %ld",i);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchIndex(char* methodName,unsigned short i) throw(ERRange)
{
    sprintf(CT.logBuffer,"No such index: %d",i);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::NoSuchCoordinate(char* methodName,char i) throw(ERRange)
{
    sprintf(CT.logBuffer,"No such coordinate: %d",i);

    CT.Error(ERR_RANGE,OH,methodName,CT.logBuffer);
}


void goblinDataObject::UnknownOption(char* methodName,int opt) throw(ERRejected)
{
    sprintf(CT.logBuffer,"Unknown option: %d",opt);

    CT.Error(ERR_REJECTED,OH,methodName,CT.logBuffer);
}


void goblinDataObject::AmountOutOfRange(char* methodName,TFloat alpha) throw(ERRejected)
{
    sprintf(CT.logBuffer,"Amount out of range: %g",alpha);

    CT.Error(ERR_REJECTED,OH,methodName,CT.logBuffer);
}


void goblinDataObject::SetLabel(char* l) throw()
{
    int newLength = 0;

    if (l!=NULL) newLength = strlen(l)+1;
    
    if (l!=NULL)
    {
        label = (char*)GoblinRealloc(label,newLength);
        strcpy(label,l);
    }
    else
    {
        delete[] label;
        label = NULL;
    }
}


void goblinDataObject::OpenFold(TModule mod,TOption opt) throw(ERRejected)
{
    CT.OpenFold(mod,opt);

    #if defined(_TIMERS_)

    CT.globalTimer[listOfModules[mod].moduleTimer] -> Enable();

    #endif
}


void goblinDataObject::CloseFold(TModule mod,TOption opt) throw(ERRejected)
{
    #if defined(_TIMERS_)

    goblinTimer* TM = CT.globalTimer[listOfModules[mod].moduleTimer];
    
    if (TM->Disable() && CT.logTimers && TM->AccTime()>0.001)
    {
        sprintf(CT.logBuffer,"Timer report (%s)",
            listOfTimers[listOfModules[mod].moduleTimer].timerName);
        LogEntry(LOG_TIMERS,CT.logBuffer);

        sprintf(CT.logBuffer,"  Cumulated times : %9.0f ms",TM->AccTime());
        LogEntry(LOG_TIMERS,CT.logBuffer);
        CT.IncreaseLogLevel();

        if (TM->PrevTime()+0.001<TM->AccTime())
        {
            sprintf(CT.logBuffer,"Previous round  : %9.0f ms",TM->PrevTime());
            LogEntry(LOG_TIMERS,CT.logBuffer);
            sprintf(CT.logBuffer,"Minimum         : %9.0f ms",TM->MinTime());
            LogEntry(LOG_TIMERS,CT.logBuffer);
            sprintf(CT.logBuffer,"Average         : %9.0f ms",TM->AvTime());
            LogEntry(LOG_TIMERS,CT.logBuffer);
            sprintf(CT.logBuffer,"Maximum         : %9.0f ms",TM->MaxTime());
            LogEntry(LOG_TIMERS,CT.logBuffer);
        }

        if (TM->FullInfo())
        {
            for (unsigned i=0;i<NoTimer;i++)
            {
                double thisTime = TM->ChildTime(TTimer(i));

                if (TTimer(i)!=listOfModules[mod].moduleTimer
                    && thisTime>0.001)
                {
                    sprintf(CT.logBuffer,"%-15s : %9.0f ms (%4.1f %%)",
                        listOfTimers[i].timerName,
                        thisTime,thisTime/TM->PrevTime()*100);
                    LogEntry(LOG_TIMERS,CT.logBuffer);
                }
            }
        }

        CT.DecreaseLogLevel();
    }

    #endif

    CT.CloseFold(mod,opt);
}


void goblinDataObject::SetLowerBound(TTimer tm,TFloat newBound)
    throw(ERRejected)
{
    #if defined(_TIMERS_)

    SetBounds(tm,newBound,InfFloat);

    #endif
}


void goblinDataObject::SetUpperBound(TTimer tm,TFloat newBound)
    throw(ERRejected)
{
    #if defined(_TIMERS_)

    SetBounds(tm,-InfFloat,newBound);

    #endif
}


void goblinDataObject::SetBounds(TTimer tm,TFloat lBound,TFloat uBound)
    throw(ERRejected)
{
    #if defined(_TIMERS_)

    #if defined(_FAILSAVE_)
             
    if (uBound<CT.globalTimer[tm]->LowerBound()-CT.epsilon)
    {
        sprintf(CT.logBuffer,"Trying to override lower bound %g with %g",
            CT.globalTimer[tm]->LowerBound(),uBound);
        Error(ERR_REJECTED,"SetBounds",CT.logBuffer);
    }

    if (lBound>CT.globalTimer[tm]->UpperBound()+CT.epsilon)
    {
        sprintf(CT.logBuffer,"Trying to override upper bound %g with %g",
            CT.globalTimer[tm]->UpperBound(),lBound);
        Error(ERR_REJECTED,"SetBounds",CT.logBuffer);
    }
             
    #endif   

    bool ret = CT.globalTimer[tm]->SetLowerBound(lBound);
    ret |= CT.globalTimer[tm]->SetUpperBound(uBound);

    if (CT.logGaps && CT.logEventHandler && ret)
    {
        if (CT.globalTimer[tm]->LowerBound()>-InfFloat)
        {
            if (CT.globalTimer[tm]->UpperBound()<InfFloat)
            {
                sprintf(CT.logBuffer,"Gap (%s) changes to [%.3f,%.3f]",
                    listOfTimers[tm].timerName,
                    CT.globalTimer[tm]->LowerBound(),
                    CT.globalTimer[tm]->UpperBound());
            }
            else
            {
                sprintf(CT.logBuffer,"Gap (%s) changes to [%.3f,*]",
                    listOfTimers[tm].timerName,
                    CT.globalTimer[tm]->LowerBound());
            }
        }
        else
        {
            sprintf(CT.logBuffer,"Gap (%s) changes to [*,%.3f]",
                listOfTimers[tm].timerName,
                CT.globalTimer[tm]->UpperBound());
        }

        LogEntry(LOG_GAPS,CT.logBuffer);
    }

    #endif
}


TFloat goblinDataObject::LowerBound(TTimer tm) throw()
{
    #if defined(_TIMERS_)

    return CT.globalTimer[tm]->LowerBound();
    
    #else
    
    return -InfFloat;

    #endif
}


TFloat goblinDataObject::UpperBound(TTimer tm) throw()
{
    #if defined(_TIMERS_)

    return CT.globalTimer[tm]->UpperBound();
    
    #else
    
    return InfFloat;

    #endif
}
