
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, August 2000.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include <display.h>


// ------------------------------------------------------


goblinDisplay::goblinDisplay(abstractMixedGraph &GC,float magnification)
    throw(ERRejected) :
    goblinDataObject(GC.Context()), G(GC), CFG(GC.Context())
{
    #if defined(_FAILSAVE_)

    if (GC.Dim()==0)
        Error(ERR_REJECTED,"goblinDisplay","Missing geometric embedding");

    #endif
    
    label = new char[255];
    
    sprintf(predColour,"#%06lX",CT.Colour(goblinController::PRED_COLOUR));
    sprintf(inftyColour,"#%06lX",CT.Colour(goblinController::INFTY_COLOUR));

    if (CFG.nodeColours==4)
    {
        maxNodeColour = 1;
        for (TNode v=0;v<G.N();v++)
        {
            TNode c = G.Colour(v);
            if (c>maxNodeColour && c!=NoNode) maxNodeColour = c;
        }
    }
    
    if (CFG.subgraph==7)
    {
        maxEdgeColour = 1;
        for (TArc a=0;a<G.M();a++)
        {
            TArc c = G.EdgeColour(2*a);
            if (c>maxEdgeColour && c!=NoArc) maxEdgeColour = c;
        }
    }

    CFG.logMeth = 0;
    CFG.xZoom = double(CFG.xZoom*magnification);
    CFG.yZoom = double(CFG.yZoom*magnification);
    CFG.xShift = int(CFG.xShift*magnification);
    CFG.yShift = int(CFG.yShift*magnification);
    CFG.nodeSize = int(CFG.nodeSize*magnification);
    CFG.arrowSize = int(CFG.arrowSize*magnification);
    
    width = (long int)(2*CFG.xShift+MaxX()*CFG.xZoom);
    height = (long int)(2*CFG.yShift+(MaxY()+CFG.legenda)*CFG.yZoom);

    LogEntry(LOG_MEM,"...Display object instanciated");
}


goblinDisplay::~goblinDisplay() throw()
{
    delete[] label;

    LogEntry(LOG_MEM,"...Display object disallocated");
}


unsigned long goblinDisplay::Allocated() throw()
{
    return 255;
}


long int goblinDisplay::AlignedCX(TNode u,TNode v) throw()
{
    if (CFG.arrows) return CanvasCX(v);

    long int dx = CanvasCX(v)-CanvasCX(u);
    long int dy = CanvasCY(v)-CanvasCY(u);
    float norm = sqrt(double(dx*dx+dy*dy));
    if (CFG.nodeStyle==0) return (long int)(CanvasCX(v)-dx*7/norm);
    else return (long int)(CanvasCX(v)-dx*CFG.nodeSize/norm);
}


long int goblinDisplay::AlignedCY(TNode u,TNode v) throw()
{
    if (CFG.arrows) return CanvasCY(v);

    long int dx = CanvasCX(v)-CanvasCX(u);
    long int dy = CanvasCY(v)-CanvasCY(u);
    float norm = sqrt(double(dx*dx+dy*dy));
    if (CFG.nodeStyle==0) return (long int)(CanvasCY(v)-dy*5/norm);
    else return (long int)(CanvasCY(v)-dy*CFG.nodeSize/norm);
};


void goblinDisplay::DisplayGraph() throw(ERRejected)
{
    for (TArc a=0;a<G.M();a++)
        if (!G.HiddenArc(2*a))  DisplayArc(a);

    for (TNode v=0;v<G.N();v++)
        if (!G.HiddenNode(v)) DisplayNode(v);
    
    if (CFG.legenda>0)
    {
        long int xm = (long int)(CFG.xShift+(MaxX()/2)*CFG.xZoom);
        long int ym = (long int)(CFG.yShift+MaxY()*CFG.yZoom+CFG.legenda*CFG.yZoom);
    
        DisplayLegenda(xm,ym,10*CFG.nodeSize);
    }
}


char* goblinDisplay::ArcLabel(TArc a,int option) throw(ERRejected)
{
    switch (option)
    {
        case 0:
        {
            strcpy(label,"");
            break;
        }
        case 1:
        {
            sprintf(label,"%ld",a);
            break;
        }
        case 2:
        {
            if (G.UCap(2*a)==InfCap) return "*";
            sprintf(label,"%g",G.UCap(2*a));
            break;
        }
        case 3:
        {
            sprintf(label,"%g",G.Sub(2*a));
            break;
        }
        case 4:
        {
            if (G.Length(2*a)==InfFloat || G.Length(2*a)==-InfFloat) return "*";
            sprintf(label,"%g",G.Length(2*a));
            break;
        }
        case 5:
        {
            sprintf(label,"%g",G.RedLength(2*a));
            break;
        }
        case 6:
        {
            sprintf(label,"%g",G.LCap(2*a));
            break;
        }
        case 7:
        {
            sprintf(label,"%ld",a+1);
            break;
        }
        case 8:
        {
            if (G.EdgeColour(2*a)==NoArc) sprintf(label,"*");
            else sprintf(label,"%ld",G.EdgeColour(2*a));
            break;
        }
        default:
        {
            UnknownOption("ArcLabel",option);
        }
    }

    return label;
}


char* goblinDisplay::NodeLabel(TNode v,int option) throw(ERRejected)
{
    switch (option)
    {
        case 0:
        {
            strcpy(label,"");
            break;
        }
        case 1:
        {
            sprintf(label,"%ld",v);
            break;
        }
        case 2:
        {
            if (G.Dist(v)==InfFloat || G.Dist(v)==-InfFloat) return "*";
            sprintf(label,"%g",G.Dist(v));
            break;
        }
        case 3:
        {
            if (G.Pi(v)==InfFloat || G.Pi(v)==-InfFloat) return "*";
            sprintf(label,"%g",G.Pi(v));
            break;
        }
        case 4:
        {
            if (G.Colour(v)==NoNode) return "*";
            sprintf(label,"%ld",G.Colour(v));
            break;
        }
        case 5:
        {
            sprintf(label,"%g",G.Demand(v));
            break;
        }
        case 6:
        {
            sprintf(label,"%ld",v+1);
            break;
        }
        default:
        {
            UnknownOption("NodeLabel",option);
        }
    }

    return label;
}


char* goblinDisplay::ArcLegenda(int option) throw(ERRejected)
{
    switch (option)
    {
        case 0:
        {
            strcpy(label,"");
            break;
        }
        case 1:
        case 7:
        {
            sprintf(label,"a");
            break;
        }
        case 2:
        {
            sprintf(label,"ucap");
            break;
        }
        case 3:
        {
            sprintf(label,"x");
            break;
        }
        case 4:
        {
            sprintf(label,"length");
            break;
        }
        case 5:
        {
            sprintf(label,"redlength");
            break;
        }
        case 6:
        {
            sprintf(label,"lcap");
            break;
        }
        case 8:
        {
            sprintf(label,"colour");
            break;
        }
        default:
        {
            UnknownOption("ArcLegenda",option);
        }
    }
    
    return label;
}


char* goblinDisplay::NodeLegenda(char* index,int option) throw(ERRejected)
{
    switch (option)
    {
        case 1:
        case 6:
        {
            sprintf(label,"%s",index);
            break;
        }
        case 2:
        {
            sprintf(label,"d(%s)",index);
            break;
        }
        case 3:
        {
            sprintf(label,"pi(%s)",index);
            break;
        }
        case 4:
        {
            sprintf(label,"colour(%s)",index);
            break;
        }
        case 5:
        {
            sprintf(label,"B(%s)",index);
            break;
        }
        default:
        {
            UnknownOption("NodeLegenda",option);
        }
    }

    return label;
}


char* goblinDisplay::FixedColour(TArc c) throw()
{
    TNode c0 = c+1;
    if (c==NoNode) c0 = 0;
    c0 += goblinController::NO_COLOUR;

    sprintf(nodeColour,"#%06lX",CT.Colour(c0));

    return nodeColour;
}


char* goblinDisplay::SmoothColour(TArc c,TNode maxColour) throw()
{
    if (c>maxColour)
    {
        sprintf(nodeColour,"#%06lX",CT.Colour(goblinController::NO_COLOUR));
        return nodeColour;
    }

    double ratio = double(c)/double(maxColour+1);
    
    long unsigned rSat = 100;
    long unsigned bSat = 100;
    long unsigned gSat = 100;

    if (ratio<0.333)
    {
        rSat = (long unsigned)(ceil(255*3*(0.333-ratio)));
        gSat = (long unsigned)(ceil(255*3*ratio));
    }
    else if (ratio<0.667)
    {
        gSat = (long unsigned)(ceil(254*3*(0.667-ratio)));
        bSat = (long unsigned)(ceil(254*3*(ratio-0.333)));
    }
    else
    {
        bSat = (long unsigned)(ceil(255*3*(1-ratio)));
        rSat = (long unsigned)(ceil(255*3*(ratio-0.667)));
    }

    sprintf(nodeColour,"#%02lX%02lX%02lX",rSat,gSat,bSat);

    return nodeColour;
}
