
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "dynq.h"
    

template <class TItem,class TKey>
dynamicQueue<TItem,TKey>::dynamicQueue(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    n = nn;
    first = NULL;
    length = 0;

    LogEntry(LOG_MEM,"...Dynamic queue instanciated");
};       


template <class TItem,class TKey>
unsigned long dynamicQueue<TItem,TKey>::Size() throw()
{
    return
          sizeof(dynamicQueue<TItem,TKey>)
        + goblinDataObject::Allocated()
        + dynamicQueue::Allocated();
};       


template <class TItem,class TKey>
unsigned long dynamicQueue<TItem,TKey>::Allocated() throw()
{
    return length*sizeof(queueMember);
};       


template <class TItem,class TKey>
char* dynamicQueue<TItem,TKey>::Display() throw()
{
    LogEntry(MSG_TRACE,"Queue");
    
    if (Empty()) LogEntry(MSG_TRACE2,"    ---");
    else
    {
        queueMember *temp;
        TItem counter = 0;

        for (temp=first; temp->next!=NULL; temp=temp->next)
        {
            if ((counter++)%10==0) LogEntry(MSG_TRACE2,"   ");

            sprintf(CT.logBuffer,"%ld, ",temp->index);
            LogEntry(MSG_APPEND,CT.logBuffer);
        }

        if (counter%10==0) LogEntry(MSG_TRACE2,"   ");

        sprintf(CT.logBuffer,"%ld (last in)",temp->index);
        LogEntry(MSG_APPEND,CT.logBuffer);
    }
    
    return NULL;
};       


template <class TItem,class TKey>
dynamicQueue<TItem,TKey>::~dynamicQueue() throw()
{
    while (!Empty()) Delete();
    
    LogEntry(LOG_MEM,"...Dynamic queue disallocated");
};


template <class TItem,class TKey>
void dynamicQueue<TItem,TKey>::Insert(TItem w,TKey alpha) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n) NoSuchItem("Insert",w);

    #endif
    
    queueMember* temp = new queueMember;
    temp -> index = w;
    temp -> next = NULL;

    if (first==NULL)
    {
        first = temp;
        last = temp;
    }
    else
    {
        last -> next = temp;
        last = temp;
    }

    length++;
};


template <class TItem,class TKey>
TItem dynamicQueue<TItem,TKey>::Delete() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Delete","Queue is empty");

    #endif

    queueMember *temp = first->next;
    TItem w = first->index;
    delete first;
    first = temp;
    length--;
    return w;
};


template <class TItem,class TKey>
TItem dynamicQueue<TItem,TKey>::Peek() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Peek","Queue is empty");

    #endif
    
    return first->index;
};


template class dynamicQueue<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class dynamicQueue<TArc,TFloat>;

#endif
