
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "dynstaq.h"
    

template <class TItem,class TKey>
dynamicStack<TItem,TKey>::dynamicStack(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    n = nn;
    top = NULL;
    depth = 0;

    LogEntry(LOG_MEM,"...Dynamic stack instanciated");
};       


template <class TItem,class TKey>
unsigned long dynamicStack<TItem,TKey>::Size() throw()
{
    return
          sizeof(dynamicStack<TItem,TKey>)
        + goblinDataObject::Allocated()
        + dynamicStack::Allocated();
};       


template <class TItem,class TKey>
unsigned long dynamicStack<TItem,TKey>::Allocated() throw()
{
    return depth*sizeof(stackMember);
};       


template <class TItem,class TKey>
char* dynamicStack<TItem,TKey>::Display() throw()
{
    LogEntry(MSG_TRACE,"Stack");
    
    if (Empty()) LogEntry(MSG_TRACE2,"    ---");
    else
    {
        stackMember *temp;
        TItem counter = 0;

        for (temp=top; temp->prev!=NULL; temp=temp->prev)
        {
            if ((counter++)%10==0) LogEntry(MSG_TRACE2,"   ");

            sprintf(CT.logBuffer,"%ld, ",temp->index);
            LogEntry(MSG_APPEND,CT.logBuffer);
        }

        if (counter%10==0) LogEntry(MSG_TRACE2,"   ");

        sprintf(CT.logBuffer,"%ld (bottom)",temp->index);
        LogEntry(MSG_APPEND,CT.logBuffer);
    }
    
    return NULL;
};       


template <class TItem,class TKey>
dynamicStack<TItem,TKey>::~dynamicStack() throw()
{
    while (!Empty()) Delete();
    
    LogEntry(LOG_MEM,"...Dynamic stack disallocated");
};


template <class TItem,class TKey>
void dynamicStack<TItem,TKey>::Insert(TItem w,TKey alpha) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n) NoSuchItem("Insert",w);

    #endif
    
    stackMember *temp = new stackMember;
    temp -> prev = top;
    temp -> index = w;
    top = temp;
    depth++;
};


template <class TItem,class TKey>
TItem dynamicStack<TItem,TKey>::Delete() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Delete","Queue is empty");

    #endif
                                                    
    stackMember *temp = top->prev;
    TItem w = top->index;
    delete top;
    top = temp;
    depth--;
    return w;
};


template <class TItem,class TKey>
TItem dynamicStack<TItem,TKey>::Peek() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Peek","Queue is empty");

    #endif
    
    return top->index;
};


template class dynamicStack<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class dynamicStack<TArc,TFloat>;

#endif
