
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, May 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "exp2tk.h"


exp2tk::exp2tk(abstractMixedGraph& GG,const char* expFileName)
    throw(ERFile) : goblinDisplay(GG,0.073), expFile(expFileName, ios::out)
{
    if (!expFile)
    {
        sprintf(CT.logBuffer,"Could not open export file %s, io_state %d",
            expFileName,expFile.rdstate());
        Error(ERR_FILE,"exp2tk",CT.logBuffer);
    }

    expFile.flags(expFile.flags() | ios::right);
    expFile.setf(ios::floatfield);
    expFile.precision(5);
    
    canvasName = "$goblinCanvas";
    smallFont = "-adobe-times-bold-r-normal--12-0-0-0-p-0-iso8859-1";
    largeFont = "-adobe-times-bold-r-normal--14-0-0-0-p-0-iso8859-1";
    
    expFile << "$goblinCanvas configure -width " << width
        << " -height " << height << endl;
    expFile << "$goblinCanvas configure -scrollregion "
        << "[list 0 0 " << width << " " << height << "]" << endl;
        
    if (strcmp(CFG.wallpaper,"")!=0)
    {
        expFile << "image create photo wallpaper -file \""
            << CFG.wallpaper << "\"" << endl;
        expFile << "$goblinCanvas create image 0 0 -image wallpaper"
            << " -anchor nw" << endl;
    }
}


unsigned long exp2tk::Size() throw()
{
    return
          sizeof(exp2tk)
        + goblinDataObject::Allocated();
}


unsigned long exp2tk::Allocated() throw()
{
    return 0;
}


void exp2tk::DisplayLegenda(long int xm,long int ym,long int radius) throw()
{
    long int xl = xm-radius;
    long int xr = xm+radius;
    
    int textShift = (int)(CFG.arrowSize);
    if (G.Blocking(1))
    {
        expFile << canvasName << " create line "
            << xl << " " << ym << " " << xr-CFG.nodeSize << " " << ym
            << " -width 2 -joinstyle bevel";

        if (CFG.arrows)
        {
            expFile << endl;
            WriteArrow(xm+CFG.arrowSize,ym,1,0);
            textShift = 2*CFG.arrowSize;
        }
        else expFile << " -arrow last -arrowshape {"
            << (int)(CFG.arrowSize*2) << " " << (int)(CFG.arrowSize*3)
            << " " << (int)(CFG.arrowSize*1) << "}" << endl;
    }
    else
    {
        expFile << canvasName << " create line "
            << xl << " " << ym << " " << xr-CFG.nodeSize << " " << ym
            << " -width 2 -joinstyle bevel" << endl;
    }
    
    if (CFG.arcLabels || CFG.arcLabels2 || CFG.arcLabels3 ||
         strcmp(CFG.arcFormatting,"")!=0)
    {
        expFile << canvasName << " create text "
            << xm << " " << (ym-textShift) << " -text {";

        if (strcmp(CFG.arcFormatting,"")==0)
        {
            char first = true;

            if (CFG.arcLabels)
            {
                expFile << ArcLegenda(CFG.arcLabels);
                first = false;
            }

            if (CFG.arcLabels2)
            {
                if (!first) expFile << "/";

                expFile << ArcLegenda(CFG.arcLabels2);
                first = false;
            }

            if (CFG.arcLabels3)
            {
                if (!first) expFile << "/";

                expFile << ArcLegenda(CFG.arcLabels3);
            }
        }
        else 
        {
            for (unsigned i=0;i<strlen(CFG.arcFormatting);i++)
            {
                if (CFG.arcFormatting[i]=='#' && i<strlen(CFG.arcFormatting)-1)
                {
                    expFile << ArcLegenda(int(CFG.arcFormatting[i+1]-'0'));
                    i++;
                    continue;
                }

                if (CFG.arcFormatting[i]!='%' || i==strlen(CFG.arcFormatting)-1)
                {
                    expFile << char(CFG.arcFormatting[i]);
                    continue;
                }

                switch (CFG.arcFormatting[i+1])
                {
                    case '1':
                    {
                        expFile << ArcLegenda(CFG.arcLabels);
                        i++;
                        break;
                    }
                    case '2':
                    {
                        expFile << ArcLegenda(CFG.arcLabels2);
                        i++;
                        break;
                    }
                    case '3':
                    {
                        expFile << ArcLegenda(CFG.arcLabels3);
                        i++;
                        break;
                    }
                    default:
                    {
                        expFile << char(CFG.arcFormatting[i]);
                    }
                }
            }
        }

        expFile << "}" << " -anchor c -font " << smallFont << endl;
    }
            
    switch (CFG.nodeStyle)
    {
        case 0:
        {
            WriteNode0(xl,ym);
            WriteNode0(xr,ym);
            WriteNodeLegenda(xl,ym+CFG.nodeSize,"u");
            WriteNodeLegenda(xr,ym+CFG.nodeSize,"v");
            
            return;
        }
        case 1:
        {
            WriteNode1(xl,ym,"#ffffff");
            WriteNode1(xr,ym,"#ffffff");
            WriteNodeLegenda(xl,ym,"u");
            WriteNodeLegenda(xr,ym,"v");
           
            return;
        }
        case 2:
        {
            WriteNode2(xl,ym,"#ffffff");
            WriteNode2(xr,ym,"#ffffff");
            WriteNodeLegenda(xl,ym,"u");
            WriteNodeLegenda(xr,ym,"v");
           
            return;
        }
    }
}


void exp2tk::WriteNodeLegenda(long int x,long int y,char* index) throw()
{
    if (strcmp(CFG.nodeFormatting,"")==0 && CFG.nodeLabels==0) return;
    
    expFile << canvasName << " create text "
        << x << " " << y
        << " -text {";

    if (strcmp(CFG.nodeFormatting,"")==0)
    {
        expFile << NodeLegenda(index,CFG.nodeLabels);
    }
    else 
    {
        for (unsigned i=0;i<strlen(CFG.nodeFormatting);i++)
        {
            if (CFG.nodeFormatting[i]=='#' && i<strlen(CFG.nodeFormatting)-1)
            {
                expFile << NodeLegenda(index,int(CFG.nodeFormatting[i+1]-'0'));
                i++;
                continue;
            }

            if (CFG.nodeFormatting[i]!='%' || i==strlen(CFG.nodeFormatting)-1)
            {
                expFile << char(CFG.nodeFormatting[i]);
                continue;
            }

            switch (CFG.nodeFormatting[i+1])
            {
                case '1':
                {
                    expFile << NodeLegenda(index,CFG.nodeLabels);
                    i++;
                    break;
                }
                default:
                {
                    expFile << char(CFG.nodeFormatting[i]);
                }
            }
        }
    }

    expFile << "}" << " -anchor c -font " << largeFont << endl;
}


void exp2tk::DisplayArc(TArc a) throw()
{
    if (G.Blocking(2*a) && G.Blocking(2*a+1)) return;
    
    if (G.HiddenArc(2*a) || G.UCap(2*a)==0) return;
    
    TNode u = G.StartNode(2*a);
    TNode v = G.EndNode(2*a);
    
    if (u==v && G.Align(2*a)==NoNode) return;
    
    bool drawPred = false;
    if (CFG.predecessors && CFG.subgraph!=0 &&
        (G.Pred(v)==2*a || G.Pred(u)==(2*a+1))
       )
    {
        drawPred = true;
        expFile << canvasName << " lower [";
        WriteArc(a,u,v,(char)0,(int)9,predColour);
        expFile << "]" << endl;
    }

    bool drawSub = false;
    if ((CFG.subgraph && (CFG.subgraph!=1 || G.Sub(2*a)>0)) ||
        (CFG.predecessors && !CFG.subgraph && (G.Pred(v)==2*a || G.Pred(u)==(2*a+1))  )
       )
    {
        drawSub = true;
        int width = 2;
        if ((CFG.subgraph==2 || CFG.subgraph==3) && G.Sub(2*a)==0) width = 1;
        if (CFG.subgraph==3 && G.Sub(2*a)==G.UCap(2*a)) width = 3;
    
        int lineStyle = 0;
        if (CFG.subgraph==2 && floor(G.Sub(2*a))!=G.Sub(2*a)) lineStyle = 1;
        if (CFG.subgraph==5)
        {
            lineStyle = int(G.Sub(2*a))%4;
            width = 1+int(G.Sub(2*a))/4;
        }
    
        if (CFG.subgraph==6)
        {
            WriteArc(a,u,v,lineStyle,width,FixedColour(G.EdgeColour(2*a)));
        }
        else if (CFG.subgraph==7)
        {
            WriteArc(a,u,v,lineStyle,width,
                SmoothColour(G.EdgeColour(2*a),maxEdgeColour));
        }
        else WriteArc(a,u,v,lineStyle,width,"#000000");

        expFile << endl;
    }

    if (drawSub && G.Orientation(2*a)!=0)
    {
        TNode y = G.Align(2*a);
        if (y!=NoNode) y = G.Interpolate(y);
        TNode x = u;
        while (y!=NoNode)
        {
            if (CFG.arrows==1 && x!=y)
            {
                if (G.Orientation(2*a)==1) DisplayArrow(a,x,y);
                else DisplayArrow(a,y,x);
            }

            x = y;
            y = G.Interpolate(y);
        }

        if (CFG.arrows && x!=v)
        {
            if (G.Orientation(2*a)==1) DisplayArrow(a,x,v);
            else DisplayArrow(a,v,x);
        }
    }

    if ((CFG.arcLabels || CFG.arcLabels2 || CFG.arcLabels3 ||
         strcmp(CFG.arcFormatting,"")) &&
        (drawPred || drawSub) && (u!=v || G.Align(2*a)!=NoNode))
        DisplayArcLabel(a,u,v);
}


void exp2tk::WriteArc(TArc a,TNode u,TNode v,int lineStyle,int width,
    char* colour) throw()
{
    TNode w = G.Align(2*a);
    if (w!=NoNode) w = G.Interpolate(w);
            
    expFile << canvasName << " create line ";
                
    if (w==NoNode)
        expFile << AlignedCX(v,u) << " " << AlignedCY(v,u) << " "
            << AlignedCX(u,v) << " " << AlignedCY(u,v) << " ";
    else
    {
        expFile << AlignedCX(w,u) << " " << AlignedCY(w,u) << " ";
        
        TNode x = w;
        while (x!=NoNode)
        {
            expFile << CanvasCX(x) << " " << CanvasCY(x) << " ";
            w = x;
            x = G.Interpolate(x);
        }
                    
        expFile << AlignedCX(w,v) << " " << AlignedCY(w,v) << " ";
    }
                
    if (CFG.arcStyle==1) expFile << " -smooth on";
    if (lineStyle==1) expFile << " -dash .";
    if (lineStyle==2) expFile << " -dash -";
    if (lineStyle==3) expFile << " -dash -.";
    
    if (!CFG.arrows && G.Orientation(2*a)==1)
        expFile << " -arrow last -arrowshape {"
            << (int)(CFG.arrowSize*2) << " " << (int)(CFG.arrowSize*3)
            << " " << (int)(CFG.arrowSize*1) << "}";

    if (!CFG.arrows && G.Orientation(2*a)==2)
        expFile << " -arrow first -arrowshape {"
            << (int)(CFG.arrowSize*2) << " " << (int)(CFG.arrowSize*3)
            << " " << (int)(CFG.arrowSize*1) << "}";

    expFile << " -joinstyle bevel -width " << width
        << " -fill " << colour;
}


void exp2tk::DisplayArrow(TArc a,TNode u,TNode v) throw()
{
    if (CFG.arrows && G.Orientation(2*a)==1)
    {
        double dx = CFG.xZoom*(CX(v)-CX(u));
        double dy = CFG.yZoom*(CY(v)-CY(u));
        double norm = sqrt(dx*dx+dy*dy);
        dx = dx/norm;
        dy = dy/norm;
    
        WriteArrow((long int)((CanvasCX(u)+CanvasCX(v))/2+CFG.arrowSize*dx),
           (long int)((CanvasCY(u)+CanvasCY(v))/2+CFG.arrowSize*dy),dx,dy);
    }

    if (CFG.arrows && G.Orientation(2*a)==2)
    {
        double dx = CFG.xZoom*(CX(u)-CX(v));
        double dy = CFG.yZoom*(CY(u)-CY(v));
        double norm = sqrt(dx*dx+dy*dy);
        dx = -dx/norm;
        dy = -dy/norm;
    
        WriteArrow((long int)((CanvasCX(u)+CanvasCX(v))/2+CFG.arrowSize*dx),
           (long int)((CanvasCY(u)+CanvasCY(v))/2+CFG.arrowSize*dy),dx,dy);
    }
}


void exp2tk::WriteArrow(long int xtop,long int ytop,double dx,double dy) throw()
{
    double ox = dy;
    double oy = -dx;
    
    expFile << canvasName << " create polygon "
        << (long int)xtop << " "
        << (long int)ytop << " "
        << (long int)(xtop-(2*dx+ox)*CFG.arrowSize) << " "
        << (long int)(ytop-(2*dy+oy)*CFG.arrowSize) << " "
        << (long int)(xtop-dx*CFG.arrowSize) << " "
        << (long int)(ytop-dy*CFG.arrowSize) << " "
        << (long int)(xtop-(2*dx-ox)*CFG.arrowSize) << " "
        << (long int)(ytop-(2*dy-oy)*CFG.arrowSize) << " "
        << (long int)xtop << " "
        << (long int)ytop
        << " -outline #000000 -fill #000000" << endl;
}


void exp2tk::DisplayArcLabel(TArc a,TNode u,TNode v) throw()
{
    long int xm = 0;
    long int ym = 0;

    if (G.Align(2*a)==NoNode)
    {
        double dx = CFG.xZoom*(CX(v)-CX(u));
        double dy = CFG.yZoom*(CY(v)-CY(u));
        double norm = sqrt(dx*dx+dy*dy);
        dx = dx/norm;
        dy = dy/norm;
    
        int textShift = (int)(CFG.arrowSize);
        if (G.Blocking(2*a+1) && CFG.arrows==1)
            textShift = (int)(2*CFG.arrowSize);
        
        xm = (long int)((CanvasCX(u)+CanvasCX(v))/2+textShift*dy);
        ym = (long int)((CanvasCY(u)+CanvasCY(v))/2-textShift*dx);
    }
    else
    {
        xm = CanvasCX(G.Align(2*a));
        ym = CanvasCY(G.Align(2*a));
    }
            
    expFile << canvasName << " create text " << xm << " " << ym << " -text {";
        
    if (strcmp(CFG.arcFormatting,"")==0)
    {
        char first = true;
        
        if (CFG.arcLabels)
        {
            expFile << ArcLabel(a,CFG.arcLabels);
            first = false;
        }

        if (CFG.arcLabels2)
        {
            if (!first) expFile << "/";

            expFile << ArcLabel(a,CFG.arcLabels2);
            first = false;
        }

        if (CFG.arcLabels3)
        {
            if (!first) expFile << "/";

            expFile << ArcLabel(a,CFG.arcLabels3);
        }
    }
    else 
    {
        for (unsigned i=0;i<strlen(CFG.arcFormatting);i++)
        {
            if (CFG.arcFormatting[i]=='#' && i<strlen(CFG.arcFormatting)-1)
            {
                expFile << ArcLabel(a,int(CFG.arcFormatting[i+1]-'0'));
                i++;
                continue;
            }

            if (CFG.arcFormatting[i]!='%' || i==strlen(CFG.arcFormatting)-1)
            {
                expFile << char(CFG.arcFormatting[i]);
                continue;
            }

            switch (CFG.arcFormatting[i+1])
            {
                case '1':
                {
                    expFile << ArcLabel(a,CFG.arcLabels);
                    i++;
                    break;
                }
                case '2':
                {
                    expFile << ArcLabel(a,CFG.arcLabels2);
                    i++;
                    break;
                }
                case '3':
                {
                    expFile << ArcLabel(a,CFG.arcLabels3);
                    i++;
                    break;
                }
                default:
                {
                    expFile << char(CFG.arcFormatting[i]);
                }
            }
        }
    }

    expFile << "} -anchor c -font " << smallFont << endl;
}


void exp2tk::DisplayNode(TNode v) throw()
{
    if (G.HiddenNode(v)) return;
    
    char* fillColour = "#ffffff";
    if (CFG.nodeColours==1 && (G.Dist(v)==InfFloat || G.Dist(v)==-InfFloat))
        fillColour = inftyColour;
    
    if (CFG.nodeColours==2) fillColour = FixedColour(G.Colour(v));
    
    if (CFG.nodeColours==3 && G.Demand(v)>0) fillColour = FixedColour(1);
    if (CFG.nodeColours==3 && G.Demand(v)<0) fillColour = FixedColour(2);

    if (CFG.nodeColours==4) fillColour = SmoothColour(G.Colour(v),maxNodeColour);
    
    long int x = CanvasCX(v);
    long int y = CanvasCY(v);
    
    switch (CFG.nodeStyle)
    {
         case 0:
         {
             WriteNode0(x,y);
             if (G.Interpolate(v)==NoNode)
                 WriteNodeLabel(v,x+CFG.nodeSize,y+CFG.nodeSize);
             else
                 WriteNodeLabel(v,CanvasCX(G.Align(v)),CanvasCY(G.Align(v)));
            
             return;
         }
         case 1:
         {
             WriteNode1(x,y,fillColour);
             WriteNodeLabel(v,x,y);
            
             return;
         }
         case 2:
         {
             WriteNode2(x,y,fillColour);
             WriteNodeLabel(v,x,y);
            
             return;
         }
    }
}


void exp2tk::WriteNode0(long int x,long int y) throw()
{
    int radius = 5;
    
    expFile << canvasName << " create oval "
        << x-radius << " " << y-radius << " "
        << x+radius << " " << y+radius << " "
        << " -outline #000000 -fill #000000 -width 1" << endl;
}


void exp2tk::WriteNode1(long int x,long int y,char* fillColour) throw()
{
    int radius = CFG.nodeSize;
    
    expFile << canvasName << " create oval "
        << x-radius << " " << y-radius << " "
        << x+radius << " " << y+radius << " "
        << " -outline #000000 -fill " << fillColour << endl;
}


void exp2tk::WriteNode2(long int x,long int y,char* fillColour) throw()
{
    int radius = CFG.nodeSize;
    
    expFile << canvasName << " create rectangle "
        << x-radius << " " << y-radius << " "
        << x+radius << " " << y+radius << " "
        << " -outline #000000 -fill " << fillColour << endl;
}


void exp2tk::WriteNodeLabel(TNode v,long int x,long int y) throw()
{
    if (strcmp(CFG.nodeFormatting,"")==0 && CFG.nodeLabels==0) return;
    
    expFile << canvasName << " create text "
        << x << " " << y
        << " -text {";

    if (strcmp(CFG.nodeFormatting,"")==0)
    {
        expFile << NodeLabel(v,CFG.nodeLabels);
    }
    else 
    {
        for (unsigned i=0;i<strlen(CFG.nodeFormatting);i++)
        {
            if (CFG.nodeFormatting[i]=='#' && i<strlen(CFG.nodeFormatting)-1)
            {
                expFile << NodeLabel(v,int(CFG.nodeFormatting[i+1]-'0'));
                i++;
                continue;
            }

            if (CFG.nodeFormatting[i]!='%' || i==strlen(CFG.nodeFormatting)-1)
            {
                expFile << char(CFG.nodeFormatting[i]);
                continue;
            }

            switch (CFG.nodeFormatting[i+1])
            {
                case '1':
                {
                    expFile << NodeLabel(v,CFG.nodeLabels);
                    i++;
                    break;
                }
                default:
                {
                    expFile << char(CFG.nodeFormatting[i]);
                }
            }
        }
    }

    expFile << "}" << " -anchor c -font " << largeFont << endl;
}


exp2tk::~exp2tk() throw()
{
    expFile << endl;
    expFile.close();
}
