
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, August 2000.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "exp2xfig.h"


exp2xfig::exp2xfig(abstractMixedGraph& GG,const char* expFileName)
    throw(ERFile) : goblinDisplay(GG,1.0), expFile(expFileName, ios::out)
{
    if (!expFile)
    {
        sprintf(CT.logBuffer,"Could not open export file %s, io_state %d",
            expFileName,expFile.rdstate());
        Error(ERR_FILE,"exp2xfig",CT.logBuffer);
    }

    expFile.flags(expFile.flags() | ios::right);
    expFile.setf(ios::floatfield);
    expFile.precision(5);
    
    expFile << "#FIG 3.2" << endl;
    expFile << "#File generated by the GOBLIN library (Release "
        << CT.majorVersion << "." << CT.minorVersion
        << CT.patchLevel << ")" << endl;

    if (MaxX()*CFG.xZoom <= MaxY()*CFG.yZoom)
        expFile << "Portrait" << endl;
    else
        expFile << "Landscape" << endl;

    expFile << "Center" << endl;
    expFile << "Metric" << endl;
    expFile << "A4" << endl;
    expFile << "100.00" << endl;
    expFile << "Single" << endl;
    expFile << "-2" << endl;
    expFile << "1200 2" << endl;
        
    // Colours
        
    expFile << "0 32 " << predColour << endl;
    expFile << "0 33 " << inftyColour << endl;
    expFile << "0 34 " << FixedColour(NoNode) << endl;
    expFile << "0 35 " << FixedColour(0) << endl;
    expFile << "0 36 " << FixedColour(1) << endl;
    expFile << "0 37 " << FixedColour(2) << endl;
    expFile << "0 38 " << FixedColour(3) << endl;
    expFile << "0 39 " << FixedColour(4) << endl;
    expFile << "0 40 " << FixedColour(5) << endl;
    expFile << "0 41 " << FixedColour(6) << endl;
    expFile << "0 42 " << FixedColour(7) << endl;
    expFile << "0 43 " << FixedColour(8) << endl;
    expFile << "0 44 " << FixedColour(9) << endl;
    expFile << "0 45 " << FixedColour(10) << endl;
    expFile << "0 46 " << FixedColour(11) << endl;
    
    // Set Magnification ratios
    
    xfigMaxX = 15000;
    xfigMaxY = 15000;
    if (MaxX()*CFG.xZoom>xfigMaxX) CFG.xZoom = int(xfigMaxX/MaxX());
    if (MaxY()*CFG.yZoom>xfigMaxY) CFG.yZoom = int(xfigMaxY/MaxY());
    
    TFloat minX = MaxX();
    TFloat minY = MaxY();
    for (TNode v=0;v<G.N();v++)
    {
        if (G.C(v,0)<minX) minX = G.C(v,0);
        if (G.C(v,1)<minY) minY = G.C(v,1);
    }
    
    // Bounding Box
    
    expFile << "2 1 0 0 0 7 150 0 -1 0.000 0 0 -1 0 0 5" << endl;
    expFile << "     "
        << int(CFG.xShift-3*CFG.nodeSize+minX*CFG.xZoom) << " "
        << int(CFG.yShift-3*CFG.nodeSize+minY*CFG.yZoom) << " "
        << int(CFG.xShift+3*CFG.nodeSize+MaxX()*CFG.xZoom) << " "
        << int(CFG.yShift-3*CFG.nodeSize+minY*CFG.yZoom) << " "
        << int(CFG.xShift+3*CFG.nodeSize+MaxX()*CFG.xZoom) << " "
        << int(CFG.yShift+3*CFG.nodeSize+(MaxY()+CFG.legenda)*CFG.yZoom) << " "
        << int(CFG.xShift-3*CFG.nodeSize+minX*CFG.xZoom) << " "
        << int(CFG.yShift+3*CFG.nodeSize+(MaxY()+CFG.legenda)*CFG.yZoom) << " "
        << int(CFG.xShift-3*CFG.nodeSize+minX*CFG.xZoom) << " "
        << int(CFG.yShift-3*CFG.nodeSize+minY*CFG.yZoom) << endl;
    
    // Merge the graph into a compound object
    
    expFile << "6 "
        << int(CFG.xShift-3*CFG.nodeSize+minX*CFG.xZoom) << " "
        << int(CFG.yShift-3*CFG.nodeSize+minY*CFG.yZoom) << " "
        << int(CFG.xShift+3*CFG.nodeSize+MaxX()*CFG.xZoom) << " "
        << int(CFG.yShift+3*CFG.nodeSize+MaxY()*CFG.yZoom)
        << endl;
}


unsigned long exp2xfig::Size() throw()
{
    return
          sizeof(exp2xfig)
        + goblinDataObject::Allocated();
}


unsigned long exp2xfig::Allocated() throw()
{
    return 0;
}


void exp2xfig::DisplayLegenda(long int xm,long int ym,long int radius) throw()
{
    int xl = xm-radius;
    int xr = xm+radius;
    int fillColour = 7;
    int forwardArrow = !CFG.arrows && G.Blocking(1);

    // Complete the graph compound object
    
    expFile << "-6" << endl;


    // Merge the legenda into a nested compound object
    
    expFile << "6 "
        << xl-CFG.nodeSize << " " << ym-CFG.nodeSize << " "
        << xr+CFG.nodeSize << " " << ym+CFG.nodeSize << " "
        << endl;
    
    expFile << "2 1 0 2 0 7 100 0 -1 0.000 0 0 -1 " << forwardArrow
        << " 0 2" << endl;
    if (forwardArrow) expFile << "2 1 3.00 120.00 150.00" << endl; 
    expFile << "     "
        << xl+CFG.nodeSize << " " << ym << " "
        << xr-CFG.nodeSize << " " << ym << " " << endl;

    int textShift = (int)(CFG.arrowSize);
    if (G.Blocking(1))
    {
        if (CFG.arrows)
        {
            WriteArrow(xm+CFG.arrowSize,ym,1,0);
            textShift = 2*CFG.arrowSize;
        }
    }
    
    if (CFG.arcLabels || CFG.arcLabels2 || CFG.arcLabels3 ||
         strcmp(CFG.arcFormatting,"")!=0)
    {
        int radius = (int)CFG.nodeSize;
        expFile << "4 1 0 30 0 2 12.000 0.000 0 "
            << 2*radius << " " << 2*radius << " "
            << xm << " ";

        expFile << ym+75-textShift << " ";

        if (strcmp(CFG.arcFormatting,"")==0)
        {
            char first = true;

            if (CFG.arcLabels)
            {
                expFile << ArcLegenda(CFG.arcLabels);
                first = false;
            }

            if (CFG.arcLabels2)
            {
                if (!first) expFile << "/";

                expFile << ArcLegenda(CFG.arcLabels2);
                first = false;
            }

            if (CFG.arcLabels3)
            {
                if (!first) expFile << "/";

                expFile << ArcLegenda(CFG.arcLabels3);
            }
        }
        else 
        {
            for (unsigned i=0;i<strlen(CFG.arcFormatting);i++)
            {
                if (CFG.arcFormatting[i]=='#' && i<strlen(CFG.arcFormatting)-1)
                {
                    expFile << ArcLegenda(int(CFG.arcFormatting[i+1]-'0'));
                    i++;
                    continue;
                }

                if (CFG.arcFormatting[i]!='%' || i==strlen(CFG.arcFormatting)-1)
                {
                    expFile << char(CFG.arcFormatting[i]);
                    continue;
                }

                switch (CFG.arcFormatting[i+1])
                {
                    case '1':
                    {
                        expFile << ArcLegenda(CFG.arcLabels);
                        i++;
                        break;
                    }
                    case '2':
                    {
                        expFile << ArcLegenda(CFG.arcLabels2);
                        i++;
                        break;
                    }
                    case '3':
                    {
                        expFile << ArcLegenda(CFG.arcLabels3);
                        i++;
                        break;
                    }
                    default:
                    {
                        expFile << char(CFG.arcFormatting[i]);
                    }
                }
            }
        }
        
        expFile << "\\001" << endl;
    }

    switch (CFG.nodeStyle)
    {
         case 0:
         {
             WriteNode0(xl,ym,fillColour);
             WriteNode0(xr,ym,fillColour);
             WriteNodeLegenda(xl,ym+CFG.nodeSize,"u");
             WriteNodeLegenda(xr,ym+CFG.nodeSize,"v");
            
             return;
         }
         case 1:
         {
             WriteNode1(xl,ym,fillColour);
             WriteNode1(xr,ym,fillColour);
             WriteNodeLegenda(xl,ym,"u");
             WriteNodeLegenda(xr,ym,"v");
            
             return;
         }
         case 2:
         {
             WriteNode2(xl,ym,fillColour);
             WriteNode2(xr,ym,fillColour);
             WriteNodeLegenda(xl,ym,"u");
             WriteNodeLegenda(xr,ym,"v");
            
             return;
         }
    }
    
    cout << "-6" << endl;
}


void exp2xfig::WriteNodeLegenda(int x,int y,char* index) throw()
{
    if (strcmp(CFG.nodeFormatting,"")==0 && CFG.nodeLabels==0) return;
    
    int radius = 100;
    expFile << "4 1 0 30 0 2 14.000 0.000 0 "
        << 2*radius << " " << 2*radius << " "
        << x << " " << y+75 << " ";
    
    if (strcmp(CFG.nodeFormatting,"")==0)
    {
        expFile << NodeLegenda(index,CFG.nodeLabels);
    }
    else 
    {
        for (unsigned i=0;i<strlen(CFG.nodeFormatting);i++)
        {
            if (CFG.nodeFormatting[i]=='#' && i<strlen(CFG.nodeFormatting)-1)
            {
                expFile << NodeLegenda(index,int(CFG.nodeFormatting[i+1]-'0'));
                i++;
                continue;
            }

            if (CFG.nodeFormatting[i]!='%' || i==strlen(CFG.nodeFormatting)-1)
            {
                expFile << char(CFG.nodeFormatting[i]);
                continue;
            }

            switch (CFG.nodeFormatting[i+1])
            {
                case '1':
                {
                    expFile << NodeLegenda(index,CFG.nodeLabels);
                    i++;
                    break;
                }
                default:
                {
                    expFile << char(CFG.nodeFormatting[i]);
                }
            }
        }
    }
    
    expFile << "\\001" << endl;
}


void exp2xfig::DisplayArc(TArc a) throw()
{
    if (G.Blocking(2*a) && G.Blocking(2*a+1)) return;
    
    if (G.HiddenArc(2*a) || G.UCap(2*a)==0) return;
    
    TNode u = G.StartNode(2*a);
    TNode v = G.EndNode(2*a);
    
    if (u==v && G.Align(2*a)==NoNode) return;
    
    bool drawPred = false;
    if (CFG.predecessors && CFG.subgraph && (G.Pred(v)==2*a || G.Pred(u)==(2*a+1)))
    {
        drawPred = true;
        WriteArc(a,u,v,(int)0,(int)9,(int)32,(int)150);
    }

    bool drawSub = false;
    if ((CFG.subgraph && (CFG.subgraph!=1 || G.Sub(2*a)>0)) ||
        (CFG.predecessors && !CFG.subgraph && (G.Pred(v)==2*a || G.Pred(u)==(2*a+1))  )
       )
    {
        drawSub = true;
        int width = 2;
        if ((CFG.subgraph==2 || CFG.subgraph==3) && G.Sub(2*a)==0) width = 1;
        if (CFG.subgraph==3 && G.Sub(2*a)==G.UCap(2*a)) width = 3;
    
        int lineStyle = 0;
        if (CFG.subgraph==2 && floor(G.Sub(2*a))!=G.Sub(2*a)) lineStyle = 1;
        if (CFG.subgraph==5)
        {
            lineStyle = int(G.EdgeColour(2*a))%4;
            width = 1+int(G.EdgeColour(2*a))/4;
        }
    
        int fillColour = 0;
        if (CFG.subgraph==6)
        {
            if (int(G.EdgeColour(2*a))>11) fillColour = 34;
            else fillColour = 35+int(G.EdgeColour(2*a))%12;
        }

        WriteArc(a,u,v,lineStyle,width,fillColour,(int)100);
    }

    if (drawSub && G.Orientation(2*a)!=0)
    {
        TNode y = G.Align(2*a);
        if (y!=NoNode) y = G.Interpolate(y);
        TNode x = u;
        while (y!=NoNode)
        {
            if (CFG.arrows==1 && x!=y)
            {
                if (G.Orientation(2*a)==1) DisplayArrow(a,x,y);
                else DisplayArrow(a,y,x);
            }
            x = y;
            y = G.Interpolate(y);
        }

        if (CFG.arrows && x!=v)
        {
            if (G.Orientation(2*a)==1) DisplayArrow(a,x,v);
            else DisplayArrow(a,v,x);
        }
    }

    if ((CFG.arcLabels || CFG.arcLabels2 || CFG.arcLabels3 ||
         strcmp(CFG.arcFormatting,"")) &&
        (drawPred || drawSub) && (u!=v || G.Align(2*a)!=NoNode))
        DisplayArcLabel(a,u,v);
}


void exp2xfig::WriteArc(TArc a,TNode u,TNode v,int lineStyle,int thickness,
    int colour,int depth) throw()
{
    float gap = 0.000;
    if (lineStyle!=0) gap = 10.000;
    int forwardArrow = !CFG.arrows && (G.Orientation(2*a)==1);
    int backwardArrow = !CFG.arrows && (G.Orientation(2*a)==2);

    TNode w = G.Align(2*a);
    if (w!=NoNode) w = G.Interpolate(w);
    TNode x = w;
    int nPoints = 2;
    while (x!=NoNode)
    {
        nPoints++;
        x = G.Interpolate(x);
    }
            
    if (CFG.arcStyle==1)    // Spline
    {
        expFile << "3 4 " << lineStyle << " " << thickness << " "
            << colour << " 7 " << depth << " 0 -1 "
            << gap << " 0 " << forwardArrow << " " << backwardArrow << " "
            << nPoints << endl;
    }
    else                // Polygon
    {
        expFile << "2 1 " << lineStyle << " " << thickness << " "
            << colour << " 7 " << depth << " 0 -1 "
            << gap << " 0 0 -1 " << forwardArrow << " " << backwardArrow << " "
            << nPoints << endl;
    }
    
    if (forwardArrow)    // Arrow
        expFile << "2 1 1.00 " << CT.arrowSize << " " << CT.arrowSize << endl; 
    if (backwardArrow)    // Arrow
        expFile << "2 1 1.00 " << CT.arrowSize << " " << CT.arrowSize << endl; 

    expFile << "     ";

    if (w==NoNode)
        expFile << AlignedCX(v,u) << " " << AlignedCY(v,u) << " "
            << AlignedCX(u,v) << " " << AlignedCY(u,v) << " ";
    else
    {
        expFile << AlignedCX(w,u) << " " << AlignedCY(w,u) << " ";
        
        x = w;
        TNode y = w;
        while (x!=NoNode)
        {
            expFile << CanvasCX(x) << " " << CanvasCY(x) << " ";
            y = x;
            x = G.Interpolate(x);
        };
                    
        expFile << AlignedCX(y,v) << " " << AlignedCY(y,v) << " ";
    }

    expFile << endl;

    if (CFG.arcStyle==1)    // Spline
    {
        expFile << "     0.000";
        x = w;

        while (x!=NoNode)
        {
            expFile << " 1.000";
            x = G.Interpolate(x);
        }

        expFile << " 0.000" << endl;
    }
}


void exp2xfig::DisplayArrow(TArc a,TNode u,TNode v) throw()
{
    if (CFG.arrows && G.Orientation(2*a)==1)
    {
        double dx = CFG.xZoom*(CX(v)-CX(u));
        double dy = CFG.yZoom*(CY(v)-CY(u));
        double norm = sqrt(dx*dx+dy*dy);
        dx = dx/norm;
        dy = dy/norm;
    
        WriteArrow((int)((CanvasCX(u)+CanvasCX(v))/2+CFG.arrowSize*dx),
           (int)((CanvasCY(u)+CanvasCY(v))/2+CFG.arrowSize*dy),dx,dy);
    }

    if (CFG.arrows && G.Orientation(2*a)==2)
    {
        double dx = CFG.xZoom*(CX(u)-CX(v));
        double dy = CFG.yZoom*(CY(u)-CY(v));
        double norm = sqrt(dx*dx+dy*dy);
        dx = -dx/norm;
        dy = -dy/norm;
    
        WriteArrow((int)((CanvasCX(u)+CanvasCX(v))/2+CFG.arrowSize*dx),
           (int)((CanvasCY(u)+CanvasCY(v))/2+CFG.arrowSize*dy),dx,dy);
    }
}


void exp2xfig::WriteArrow(int xtop,int ytop,double dx,double dy) throw()
{
    double ox = dy;
    double oy = -dx;
    
    expFile << "2 3 0 1 0 0 100 0 20 0.000 1 0 -1 0 0 5" << endl;
    expFile << "     "
        << (int)xtop << " "
        << (int)ytop << " "
        << (int)(xtop-(1.8*dx+0.7*ox)*CFG.arrowSize) << " "
        << (int)(ytop-(1.8*dy+0.7*oy)*CFG.arrowSize) << " "
        << (int)(xtop-1.4*dx*CFG.arrowSize) << " "
        << (int)(ytop-1.4*dy*CFG.arrowSize) << " "
        << (int)(xtop-(1.8*dx-0.7*ox)*CFG.arrowSize) << " "
        << (int)(ytop-(1.8*dy-0.7*oy)*CFG.arrowSize) << " "
        << (int)xtop << " "
        << (int)ytop << endl;
};


void exp2xfig::DisplayArcLabel(TArc a,TNode u,TNode v) throw()
{
    int xm = 0;
    int ym = 0;
    if (G.Align(2*a)==NoNode)
    {
        double dx = CFG.xZoom*(CX(v)-CX(u));
        double dy = CFG.yZoom*(CY(v)-CY(u));
        double norm = sqrt(dx*dx+dy*dy);
        dx = dx/norm;
        dy = dy/norm;
    
        int textShift = (int)(CFG.arrowSize);
        if (G.Blocking(2*a+1) && CFG.arrows)
            textShift = (int)(2*CFG.arrowSize);
        
        xm = (int)((CanvasCX(u)+CanvasCX(v))/2+textShift*dy);
        ym = (int)((CanvasCY(u)+CanvasCY(v))/2-textShift*dx);
    }
    else
    {
        xm = (int)CanvasCX(G.Align(2*a));
        ym = (int)CanvasCY(G.Align(2*a));
    }
            
    int radius = (int)CFG.nodeSize;
        
    expFile << "4 1 0 30 0 2 12.000 0.000 0 "
        << 2*radius << " " << 2*radius << " "
        << xm << " " << ym+75 << " ";

    if (strcmp(CFG.arcFormatting,"")==0)
    {
        char first = true;
        
        if (CFG.arcLabels)
        {
            expFile << ArcLabel(a,CFG.arcLabels);
            first = false;
        }

        if (CFG.arcLabels2)
        {
            if (!first) expFile << "/";

            expFile << ArcLabel(a,CFG.arcLabels2);
            first = false;
        }

        if (CFG.arcLabels3)
        {
            if (!first) expFile << "/";

            expFile << ArcLabel(a,CFG.arcLabels3);
        }
    }
    else 
    {
        for (unsigned i=0;i<strlen(CFG.arcFormatting);i++)
        {
            if (CFG.arcFormatting[i]=='#' && i<strlen(CFG.arcFormatting)-1)
            {
                expFile << ArcLabel(a,int(CFG.arcFormatting[i+1]-'0'));
                i++;
                continue;
            }

            if (CFG.arcFormatting[i]!='%' || i==strlen(CFG.arcFormatting)-1)
            {
                expFile << char(CFG.arcFormatting[i]);
                continue;
            }

            switch (CFG.arcFormatting[i+1])
            {
                case '1':
                {
                    expFile << ArcLabel(a,CFG.arcLabels);
                    i++;
                    break;
                }
                case '2':
                {
                    expFile << ArcLabel(a,CFG.arcLabels2);
                    i++;
                    break;
                }
                case '3':
                {
                    expFile << ArcLabel(a,CFG.arcLabels3);
                    i++;
                    break;
                }
                default:
                {
                    expFile << char(CFG.arcFormatting[i]);
                }
            }
        }
    }

    expFile << "\\001" << endl;
}


void exp2xfig::DisplayNode(TNode v) throw()
{
    if (G.HiddenNode(v)) return;
    
    int fillColour = 7;
    if (CFG.nodeColours==1 && (G.Dist(v)==InfFloat || G.Dist(v)==-InfFloat))
        fillColour = 33;
    
    if (CFG.nodeColours==2 && G.Colour(v)>11)
        fillColour = 34;
    if (CFG.nodeColours==2 && G.Colour(v)<12)
        fillColour = 35+G.Colour(v)%12;
    
    if (CFG.nodeColours==3 && G.Demand(v)>0) fillColour = 36;
    if (CFG.nodeColours==3 && G.Demand(v)<0) fillColour = 37;
    
    int x = CanvasCX(v);
    int y = CanvasCY(v);
    
    switch (CFG.nodeStyle)
    {
         case 0:
         {
             WriteNode0(x,y,fillColour);
             if (G.Interpolate(v)==NoNode)
                 WriteNodeLabel(v,x+CFG.nodeSize,y+CFG.nodeSize);
             else
                 WriteNodeLabel(v,CanvasCX(G.Align(v)),
                                  CanvasCY(G.Align(v)));
            
             return;
         }
         case 1:
         {
             WriteNode1(x,y,fillColour);
             WriteNodeLabel(v,x,y);
            
             return;
         }
         case 2:
         {
             WriteNode2(x,y,fillColour);
             WriteNodeLabel(v,x,y);
            
             return;
         }
    }
}


void exp2xfig::WriteNode0(int x,int y,int fillColour) throw()
{
    int radius = 60;
            
    expFile << "1 3 0 4 0 0 50 0 20 0.000 1 0.0000 "
        << x << " " << y << " "
        << radius << " " << radius << " "
        << x-radius << " " << y << " "
        << x+radius  << " " << y << endl;
}


void exp2xfig::WriteNode1(int x,int y,int fillColour) throw()
{
    int radius = (int)CFG.nodeSize;
            
    expFile << "1 3 0 1 0 " << fillColour
        << " 50 0 20 0.000 1 0.0000 "
        << x << " " << y << " "
        << radius << " " << radius << " "
        << x-radius << " " << y << " "
        << x+radius << " " << y << endl;
}


void exp2xfig::WriteNode2(int x,int y,int fillColour) throw()
{
    int radius = (int)CFG.nodeSize;
            
    expFile << "2 2 0 1 0 " << fillColour
        << " 50 0 20 0.000 0 0 0 0 0 5 " << endl
        << x-radius << " " << y-radius << " "
        << x-radius << " " << y+radius << " "
        << x+radius << " " << y+radius << " "
        << x+radius << " " << y-radius << " "
        << x-radius << " " << y-radius << " " << endl;
}


void exp2xfig::WriteNodeLabel(TNode v,int x,int y) throw()
{
    if (strcmp(CFG.nodeFormatting,"")==0 && CFG.nodeLabels==0) return;
    
    int radius = 100;
    expFile << "4 1 0 30 0 2 14.000 0.000 0 "
        << 2*radius << " " << 2*radius << " "
        << x << " " << y+75 << " ";
    
    if (strcmp(CFG.nodeFormatting,"")==0)
    {
        expFile << NodeLabel(v,CFG.nodeLabels);
    }
    else 
    {
        for (unsigned i=0;i<strlen(CFG.nodeFormatting);i++)
        {
            if (CFG.nodeFormatting[i]=='#' && i<strlen(CFG.nodeFormatting)-1)
            {
                expFile << NodeLabel(v,int(CFG.nodeFormatting[i+1]-'0'));
                i++;
                continue;
            }

            if (CFG.nodeFormatting[i]!='%' || i==strlen(CFG.nodeFormatting)-1)
            {
                expFile << char(CFG.nodeFormatting[i]);
                continue;
            }

            switch (CFG.nodeFormatting[i+1])
            {
                case '1':
                {
                    expFile << NodeLabel(v,CFG.nodeLabels);
                    i++;
                    break;
                }
                default:
                {
                    expFile << char(CFG.nodeFormatting[i]);
                }
            }
        }
    }
    
    expFile << "\\001" << endl;
}


exp2xfig::~exp2xfig() throw()
{
    expFile << endl;
    expFile.close();
}
