
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, April 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "globals.h"
#include <malloc.h>


#if defined(_BIG_NODES_)

TNode     NoNode   = 200000;

#else

TNode     NoNode   = 50000;

#endif


#if defined(_SMALL_ARCS_)

TArc      NoArc    = 50000;

#else

TArc      NoArc    = 2000000000;

#endif


TCap      InfCap   = 1000000000;
TFloat    InfFloat = 1000000000;
THandle   NoHandle = 2000000000;
TIndex    NoIndex  = 2000000000;
TVar      NoVar    = 2000000000;
TRestr    NoRestr  = 2000000000;


long unsigned goblinHeapSize   = 0;
long unsigned goblinMaxSize   = 0;
long unsigned goblinNFragments = 0;
long unsigned goblinNAllocs   = 0;
long unsigned goblinNObjects  = 0;


#if defined(_HEAP_MON_)

void* operator new(size_t size) throw (std::bad_alloc)
{
    if (size==0) return (void*)(0);
    
    if (void* p = malloc(size+sizeof(size_t)))
    {
        goblinHeapSize += size;

        if (goblinHeapSize>goblinMaxSize) goblinMaxSize = goblinHeapSize;

        goblinNFragments++;
        goblinNAllocs++;

        *((size_t*)(p)) = size;
        return (size_t*)(p)+1;
    }

    throw std::bad_alloc();
}


void* operator new[](size_t size) throw (std::bad_alloc)
{
    if (size==0) return (void*)(0);

    if (void* p = malloc(size+sizeof(size_t)))
    {
        goblinHeapSize += size;

        if (goblinHeapSize>goblinMaxSize) goblinMaxSize = goblinHeapSize;

        goblinNFragments++;
        goblinNAllocs++;

        *((size_t*)(p)) = size;
        return (size_t*)(p)+1;
    }
    
    throw std::bad_alloc();
}


void* GoblinRealloc(void* p,size_t size) throw (std::bad_alloc)
{
    if (p==NULL)
    {
        if (size==0) return NULL;
        
        if (void* p = malloc(size+sizeof(size_t)))
        {
            goblinHeapSize += size;

            if (goblinHeapSize>goblinMaxSize) goblinMaxSize = goblinHeapSize;

            goblinNFragments++;
            goblinNAllocs++;

            *((size_t*)(p)) = size;
            return (size_t*)(p)+1;
        }

        throw std::bad_alloc();
    }
    else
    {
        size_t old_size = *((size_t*)(p)-1);

        if (size==old_size) return p;

        goblinHeapSize += size-old_size;

        if (size==0)
        {
            goblinNFragments--;
            free((size_t*)(p)-1);
            return NULL;
        }

        if (size_t* q = (size_t*)malloc(size+sizeof(size_t)))
        {
            if (goblinHeapSize>goblinMaxSize) goblinMaxSize = goblinHeapSize;

            goblinNAllocs++;

            *q = size;

            if (size>old_size)
                memcpy(q+1,p,old_size);
            else memcpy(q+1,p,size);

            free((size_t*)(p)-1);

            return q+1;
        }
    
        throw std::bad_alloc();
    }
    
    throw std::bad_alloc();
}


void operator delete(void* p) throw ()
{
    if (p==NULL) return;
    
    size_t size = *((size_t*)(p)-1);
    goblinHeapSize -= size;
    goblinNFragments--;

    free((size_t*)(p)-1);
}


void operator delete[](void* p) throw ()
{
    if (p==NULL) return;

    size_t size = *((size_t*)(p)-1);

    goblinHeapSize -= size;
    goblinNFragments--;

    free((size_t*)(p)-1);
}

#else

void* GoblinRealloc(void* p,size_t size) throw (std::bad_alloc)
{
    return realloc(p,size);
}

#endif
