
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, January 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "auxnet.h"


iLayeredAuxNetwork::iLayeredAuxNetwork(layeredAuxNetwork &GC) throw() : G(GC)
{                                                   
    n = G.n;
    currentIndex = new TArc[n];

    for (TNode i=0;i<n;i++) currentIndex[i] = 0;
};

     
unsigned long iLayeredAuxNetwork::Size() throw()
{   
    return
          sizeof(iLayeredAuxNetwork)
        + goblinDataObject::Allocated()
        + iLayeredAuxNetwork::Allocated();
};


unsigned long iLayeredAuxNetwork::Allocated() throw()
{   
    return n*sizeof(TArc);        // currentIndex[]
};


void iLayeredAuxNetwork::Reset() throw()
{
    for (TNode i=0;i<n;i++) currentIndex[i] = 0;
}                                              


void iLayeredAuxNetwork::Reset(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Reset",v);

    #endif
    
    currentIndex[v] = 0;
}                                              


TArc iLayeredAuxNetwork::Read(TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Read",v);
    
    if (currentIndex[v]>=G.inDegree[v]) NoMoreArcs("Read",v);

    #endif

    return G.prop[v][currentIndex[v]++];
}


TArc iLayeredAuxNetwork::Peek(TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Peek",v);
    
    if (currentIndex[v]>=G.inDegree[v]) NoMoreArcs("Peek",v);

    #endif

    return G.prop[v][currentIndex[v]];
}


void iLayeredAuxNetwork::Skip(TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Skip",v);
    
    if (currentIndex[v]>=G.inDegree[v]) NoMoreArcs("Skip",v);

    #endif

    currentIndex[v]++;
}


bool iLayeredAuxNetwork::Active(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Active",v);

    #endif
    
    return currentIndex[v]<G.inDegree[v];
};

     
iLayeredAuxNetwork::~iLayeredAuxNetwork() throw()
{                                                   
    delete[] currentIndex;
};
