
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, April 2004.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "ilpWrapper.h"
#include "export.h"


// -----------------------------------------------------------------------------


goblinILPWrapper* goblinILPWrapper::DualForm() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    for (TVar i=0;i<L();i++)
        if (VarType(i)==VAR_INT)
            Error(ERR_REJECTED,"DualForm","Integer variable found");

    #endif
    

    goblinILPModule* ModLP = (goblinILPModule*)CT.pLPModule;
    goblinILPWrapper* XLP  =
        ModLP->NewInstance(L(),2*K(),NZ(),MAXIMIZE,CT);
    
    TFloat sign = 1;
    
    if (Direction()==MAXIMIZE)
    {
        XLP->SetDirection(MINIMIZE);
        sign = -1;
    }


    TVar j = 0;

    TIndex* index = new TIndex[L()];
    TFloat* val   = new TFloat[L()];

    for (TVar i=0;i<L();i++)
    {        
        // Transform variables to structural restrictions

        XLP -> AddRestr(Cost(i),Cost(i));

        if (strlen(VarLabel(i))+3>bufferLength)
        {
            bufferLength = strlen(VarLabel(i))+3;
            labelBuffer = (char*)GoblinRealloc(labelBuffer,bufferLength);
        }

        sprintf(labelBuffer,"%s",VarLabel(i));
        XLP -> SetRestrLabel(i,labelBuffer,OWNED_BY_SENDER);
        
        if (Direction()==MINIMIZE)
        {
            if (URange(i)==0)
                XLP -> SetUBound(i,InfFloat);
            else if (LRange(i)==0)
                XLP -> SetLBound(i,-InfFloat);
        }
        else
        {
            if (URange(i)==0)
                XLP -> SetLBound(i,-InfFloat);
            else if (LRange(i)==0)
                XLP -> SetUBound(i,InfFloat);
        }
    }


    for (TRestr i=0;i<K();i++)
    {
        // Transform structural restrictions to variables

        if (UBound(i)==InfFloat && LBound(i)==-InfFloat) continue;
        
        TIndex nz = GetRow(i,index,val);
        
        if (strlen(RestrLabel(i))+2>bufferLength)
        {
            bufferLength = strlen(RestrLabel(i))+2;
            labelBuffer = (char*)GoblinRealloc(labelBuffer,bufferLength);
        }

        if (LBound(i)>-InfFloat && LBound(i)<UBound(i))
        {
            if (Direction()==MINIMIZE)
                XLP -> AddVar(0,InfFloat,LBound(i));
            else
                XLP -> AddVar(-InfFloat,0,LBound(i));
            
            XLP -> SetColumn(j,nz,index,val);
            
            if (UBound(i)<InfFloat)
                sprintf(labelBuffer,"%sl",RestrLabel(i));
            else sprintf(labelBuffer,"%s",RestrLabel(i));

            XLP -> SetVarLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
        
        if (UBound(i)<InfFloat)
        {
            if (LBound(i)<UBound(i))
            {
                if (Direction()==MINIMIZE)
                    XLP -> AddVar(-InfFloat,0,UBound(i));
                else
                    XLP -> AddVar(0,InfFloat,UBound(i));
            }
            else
            {
                XLP -> AddVar(-InfFloat,InfFloat,UBound(i));
            }
            
            XLP -> SetColumn(j,nz,index,val);

            if (LBound(i)>-InfFloat && LBound(i)<UBound(i))
                sprintf(labelBuffer,"%su",RestrLabel(i));
            else sprintf(labelBuffer,"%s",RestrLabel(i));

            XLP -> SetVarLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
    }


    for (TVar i=0;i<L();i++)
    {        
        // Transform variable range restrictions

        if (URange(i)==InfFloat && LRange(i)==-InfFloat) continue;
        if (LRange(i)==-InfFloat && URange(i)==0) continue;
        if (URange(i)==InfFloat && LRange(i)==0) continue;
        
        index[0] = i;
        val[0] = 1;
        
        if (LRange(i)>-InfFloat && (LRange(i)!=0 || LRange(i)==URange(i)))
        {
            if (Direction()==MINIMIZE)
                XLP -> AddVar(0,InfFloat,LRange(i));
            else
                XLP -> AddVar(-InfFloat,0,LRange(i));
            
            XLP -> SetColumn(j,1,index,val);
            
            if (URange(i)<InfFloat && URange(i)!=0)
                sprintf(labelBuffer,"a%sl",VarLabel(i));
            else sprintf(labelBuffer,"a%s",VarLabel(i));

            XLP -> SetVarLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
        
        if (URange(i)<InfFloat && URange(i)!=0)
        {
            if (LRange(i)<URange(i))
            {
                if (Direction()==MINIMIZE)
                    XLP -> AddVar(-InfFloat,0,URange(i));
                else
                    XLP -> AddVar(0,InfFloat,URange(i));
            }
            else
            {
                XLP -> AddVar(-InfFloat,InfFloat,URange(i));
            }
            
            XLP -> SetColumn(j,1,index,val);

            if (LRange(i)>-InfFloat && (LRange(i)!=0 || LRange(i)==URange(i)))
                sprintf(labelBuffer,"a%su",VarLabel(i));
            else sprintf(labelBuffer,"a%s",VarLabel(i));

            XLP -> SetVarLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
    }
    
    XLP -> ResetBasis();


    delete[] index;
    delete[] val;
    
    return XLP;
}


goblinILPWrapper* goblinILPWrapper::StandardForm() throw(ERRejected)
{
    goblinILPModule* ModLP = (goblinILPModule*)CT.pLPModule;
    goblinILPWrapper* XLP  =
        ModLP->NewInstance(2*K()+L()+1,2*K()+L()+1,2*(NZ()+L()),Direction(),CT);

    TVar* index = new TVar[L()];
    TFloat* val = new TFloat[L()];
    TVar* map = new TVar[2*L()];


    // Transform original variables, ranges and objective function

    TRestr k = 0;       // Number of restrictions in the transformed LP
    TVar l = 0;         // Number of variable in the transformed LP
    TFloat offset = 0;  // Constant to achieve the same optimum in both LPs

    for (TVar i=0;i<L();i++)
    {
        if (LRange(i)==URange(i))
        {
            // Variable can be omitted

            offset += Cost(i)*LRange(i);

            map[2*i] = map[2*i+1] = NoVar;

            continue;
        }

        if (strlen(VarLabel(i))+3>bufferLength)
        {
            bufferLength = strlen(VarLabel(i))+3;
            labelBuffer = (char*)GoblinRealloc(labelBuffer,bufferLength);
        }

        if (LRange(i)>-InfFloat)
        {
            XLP -> AddVar(0,InfFloat,Cost(i),VarType(i));
            
            sprintf(labelBuffer,"%s",VarLabel(i));
            XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);

            offset += Cost(i)*LRange(i);
            
            map[2*i] = l;
            map[2*i+1] = NoVar;
            
            if (URange(i)<InfFloat)
            {
                // Add explicit restriction for upper variable bound
            
                XLP -> AddRestr(URange(i)-LRange(i),URange(i)-LRange(i));
                XLP -> SetCoeff(k,l,1);
            
                sprintf(labelBuffer,"r%s",VarLabel(i));
                XLP -> SetRestrLabel(k,labelBuffer,OWNED_BY_SENDER);
                
                // Add slack variable
                
                l++;

                XLP -> AddVar(0,InfFloat,0,VAR_FLOAT);
                XLP -> SetCoeff(k,l,1);

                sprintf(labelBuffer,"s%sp",VarLabel(i));
                XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);

                k++;
            }

            l++;
        }
        else
        {
            if (URange(i)<InfFloat)
            {
                XLP -> AddVar(0,InfFloat,-Cost(i),VarType(i));

                sprintf(labelBuffer,"%s",VarLabel(i));
                XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);
                
                offset += Cost(i)*URange(i);
            
                map[2*i] = NoVar;
                map[2*i+1] = l;

                l++;
            }
            else
            {
                // Requires two non-negative variables
            
                XLP -> AddVar(0,InfFloat,Cost(i),VarType(i));

                sprintf(labelBuffer,"%sp",VarLabel(i));
                XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);
            
                map[2*i] = l;
                
                l++;

                XLP -> AddVar(0,InfFloat,-Cost(i),VarType(i));

                sprintf(labelBuffer,"%sn",VarLabel(i));
                XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);
            
                map[2*i+1] = l;
                
                l++;
            }
        }
    }


    // Add structural restrictions and respective slack variables

    for (TRestr i=0;i<K();i++)
    {
        if (UBound(i)==InfFloat && LBound(i)==-InfFloat) continue;

        // Determine right hand side

        TIndex nz = GetRow(i,index,val);
        TFloat RHS = 0;

        for (TIndex j=0;j<nz;j++)
        {
            if (LRange(index[j])>-InfFloat)
                RHS += LRange(index[j])*val[j];
            else if (URange(index[j])<InfFloat)
                RHS += URange(index[j])*val[j];
        }
    
        if (strlen(RestrLabel(i))+3>bufferLength)
        {
            bufferLength = strlen(RestrLabel(i))+3;
            labelBuffer = (char*)GoblinRealloc(labelBuffer,bufferLength);
        }

        if (UBound(i)>LBound(i))
        {
            if (UBound(i)<InfFloat)
            {
                // Replace "less or equal" condition by using slack variable
            
                XLP -> AddVar(0,InfFloat,0,VAR_FLOAT);

                sprintf(labelBuffer,"s%sp",RestrLabel(i));
                XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);

                // Add equality restriction
            
                XLP -> AddRestr(UBound(i)-RHS,UBound(i)-RHS);
            
                sprintf(labelBuffer,"%su",RestrLabel(i));
                XLP -> SetRestrLabel(k,labelBuffer,OWNED_BY_SENDER);
                
                XLP -> SetCoeff(k,l,1);

                // Copy matrix coefficients

                for (TIndex j=0;j<nz;j++)
                {
                    if (map[2*index[j]]!=NoVar)
                        XLP -> SetCoeff(k,map[2*index[j]],val[j]);

                    if (map[2*index[j]+1]!=NoVar)
                        XLP -> SetCoeff(k,map[2*index[j]+1],-val[j]);
                }

                k++;
                l++;
            }

            if (LBound(i)>-InfFloat)
            {
                // Replace "greater or equal" condition using a slack variable
            
                XLP -> AddVar(0,InfFloat,0,VAR_FLOAT);

                sprintf(labelBuffer,"s%sn",RestrLabel(i));
                XLP -> SetVarLabel(l,labelBuffer,OWNED_BY_SENDER);
                
                // Add equality restriction
            
                XLP -> AddRestr(LBound(i)-RHS,LBound(i)-RHS);
            
                sprintf(labelBuffer,"%sl",RestrLabel(i));
                XLP -> SetRestrLabel(k,labelBuffer,OWNED_BY_SENDER);
                
                XLP -> SetCoeff(k,l,-1);

                // Copy matrix coefficients

                for (TIndex j=0;j<nz;j++)
                {
                    if (map[2*index[j]]!=NoVar)
                        XLP -> SetCoeff(k,map[2*index[j]],val[j]);

                    if (map[2*index[j]+1]!=NoVar)
                        XLP -> SetCoeff(k,map[2*index[j]+1],-val[j]);
                }

                k++;
                l++;
            }
        }
        else
        {
            // Already an equality restriction

            XLP -> AddRestr(LBound(i)-RHS,LBound(i)-RHS);

            sprintf(labelBuffer,"%s",RestrLabel(i));
            XLP -> SetRestrLabel(k,labelBuffer,OWNED_BY_SENDER);

            // Copy matrix coefficients

            for (TIndex j=0;j<nz;j++)
            {
                if (map[2*index[j]]!=NoVar)
                    XLP -> SetCoeff(k,map[2*index[j]],val[j]);

                if (map[2*index[j]+1]!=NoVar)
                    XLP -> SetCoeff(k,map[2*index[j]+1],-val[j]);
            }

            k++;
        }
    }


    // Add dummy variable for the constant offset

    TFloat sign = 1;
    if (offset<0) sign = -1;
    
    XLP -> AddVar(0,InfFloat,sign,VAR_FLOAT);
    XLP -> SetVarLabel(l,"xofs",OWNED_BY_SENDER);

    XLP -> AddRestr(offset,offset);
    XLP -> SetRestrLabel(k,"rofs",OWNED_BY_SENDER);

    XLP -> SetCoeff(k,l,sign);
        
    k++;
    l++;

    
    XLP -> ResetBasis();

    
    delete[] index;
    delete[] val;
    delete[] map;

    return XLP;
}


goblinILPWrapper* goblinILPWrapper::CanonicalForm() throw(ERRejected)
{
    goblinILPModule* ModLP = (goblinILPModule*)CT.pLPModule;
    goblinILPWrapper* XLP  =
        ModLP->NewInstance(2*(K()+L()),L(),2*(NZ()+L()),Direction(),CT);

    TVar* index = new TVar[L()];
    TFloat* val = new TFloat[L()];
    TRestr j = 0;

    // Transform variables and objective function

    for (TVar i=0;i<L();i++)
    {        
        XLP -> AddVar(-InfFloat,InfFloat,Cost(i),VarType(i));
            
        if (strlen(VarLabel(i))+2>bufferLength)
        {
            bufferLength = strlen(VarLabel(i))+2;
            labelBuffer = (char*)GoblinRealloc(labelBuffer,bufferLength);
        }

        sprintf(labelBuffer,"%s",VarLabel(i));
        XLP -> SetVarLabel(i,labelBuffer,OWNED_BY_SENDER);
    }

    // Transform structural restrictions

    for (TRestr i=0;i<K();i++)
    {
        if (UBound(i)==InfFloat && LBound(i)==-InfFloat) continue;
    
        TIndex nz = GetRow(i,index,val);
        
        if (strlen(RestrLabel(i))+2>bufferLength)
        {
            bufferLength = strlen(RestrLabel(i))+2;
            labelBuffer = (char*)GoblinRealloc(labelBuffer,bufferLength);
        }

        if (UBound(i)<InfFloat)
        {
            XLP -> AddRestr(-InfFloat,UBound(i));
            XLP -> SetRow(j,nz,index,val);
            
            if (LBound(i)>-InfFloat)
                sprintf(labelBuffer,"%su",RestrLabel(i));
            else sprintf(labelBuffer,"%s",RestrLabel(i));
            XLP -> SetRestrLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
        
        if (LBound(i)>-InfFloat)
        {
            for (TVar k=0;k<nz;k++) val[k] *= -1;
        
            XLP -> AddRestr(-InfFloat,-LBound(i));
            XLP -> SetRow(j,nz,index,val);
            
            sprintf(labelBuffer,"%sl",RestrLabel(i));
            XLP -> SetRestrLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
    }
    
    // Transform variable range restrictions

    for (TVar i=0;i<L();i++)
    {
        if (URange(i)==InfFloat && LRange(i)==-InfFloat) continue;
    
        if (URange(i)<InfFloat)
        {
            XLP -> AddRestr(-InfFloat,URange(i));
            XLP -> SetCoeff(j,i,1);
            
            sprintf(labelBuffer,"%su",VarLabel(i));
            XLP -> SetRestrLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
        
        if (LRange(i)>-InfFloat)
        {
            XLP -> AddRestr(-InfFloat,-LRange(i));
            XLP -> SetCoeff(j,i,-1);
            
            sprintf(labelBuffer,"%sl",VarLabel(i));
            XLP -> SetRestrLabel(j,labelBuffer,OWNED_BY_SENDER);

            j++;
        }
    }
    
    XLP -> ResetBasis();
    
    delete[] index;
    delete[] val;

    return XLP;
}


void goblinILPWrapper::FlipDirection() throw()
{
    if (Direction()==MAXIMIZE)
    {
        SetDirection(MINIMIZE);
    }
    else if (Direction()==MINIMIZE)
    {
        SetDirection(MAXIMIZE);
    }
    
    for (TVar i=0;i<L();i++) SetCost(i,-Cost(i));
}
