
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, August 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absbal.h"
#include "bal2bal.h"
#include "surgra.h"


TFloat abstractBalancedFNW::MinCBalFlow(TNode s) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("MinCBalFlow",s);

    if (MaxLCap()>0)
        Error(ERR_REJECTED,"MinCBalFlow","Non-trivial lower bounds");

    #endif

    OpenFold(ModMinCBalFlow,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing minimum cost balanced flow...");

    TFloat ret = InfFloat;
    
    switch (CT.methMinCBalFlow)
    {
        case 0:
        {
            ret = PrimalDual(s);
            break;
        }
        case 1:
        {
            ret = EnhancedPD(s);
            break;
        }
        default:
        {
            UnknownOption("MinCBalFlow",CT.methMinCBalFlow);
            CloseFold(ModMinCBalFlow,NO_INDENT);
            throw ERRejected();
        }
    }

    CloseFold(ModMinCBalFlow,NO_INDENT);

    return ret;
}


TFloat abstractBalancedFNW::PrimalDual(TNode s) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("PrimalDual",s);

    if (MaxLCap()>0) Error(ERR_REJECTED,"PrimalDual","Non-trivial lower bounds");

    #endif
    
    OpenFold(ModPrimalDual,SHOW_TITLE);
    
    CT.Ping(Handle(),n*m);
    
    TFloat ret = Weight();
    SetLowerBound(TimerPrimalDual,ret);

    surfaceGraph G(*this);
    
    switch (CT.methPrimalDual)
    {
        case 0:
        {
            G.PrimalDual0(s);
            break;
        }
        case 1:
        case 2:
        {
            G.PrimalDual1(s);
            break;
        }
        default:
        {
            UnknownOption("PrimalDual",CT.methPrimalDual);
        }
    }
    
    CloseFold(ModPrimalDual);
    
    #if defined(_FAILSAVE_)

    if (CT.methFailSave==1 && !G.Compatible())
    {
        InternalError("PrimalDual","Solutions are corrupted");
    }

    #endif

    ret = Weight();

    SetBounds(TimerPrimalDual,ret,ret);
    
    return ret;
}


TFloat abstractBalancedFNW::EnhancedPD(TNode s) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("EnhancedPD",s);

    if (MaxLCap()>0)
        Error(ERR_REJECTED,"EnhancedPD","Non-trivial lower bounds");

    #endif
    
    OpenFold(ModEnhancedPD,SHOW_TITLE);

    MinCostFlow(s,s^1);
    if (pi==NULL)
    {
        LogEntry(LOG_METH,"Computing optimal node potentials...");
        OpenFold();
        CycleCanceling();
        CloseFold();
    }
    
    CancelEven();
    TFloat ret = CancelPD();

    CloseFold(ModEnhancedPD);
    
    return ret;
}

 
TFloat abstractBalancedFNW::CancelPD() throw()
{
    #if defined(_FAILSAVE_)

    if (Q==NULL) Error(ERR_REJECTED,"CancelPD","No odd cycles present");

    #endif
    
    bal2bal G(*this);

    LogEntry(LOG_METH,"Refining balanced flow...");

    return G.PrimalDual(G.Source());
}
