
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "statq.h"
    

template <class TItem,class TKey>
staticQueue<TItem,TKey>::staticQueue(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    n = nn;
    next = new TItem[n];
    first = n;
    master = true;
    length = 0;
    
    for (TItem i=0;i<n;i++) next[i] = n;

    LogEntry(LOG_MEM,"...Static queue allocated");
}


template <class TItem,class TKey>
staticQueue<TItem,TKey>::staticQueue(staticQueue<TItem,TKey> &Q) throw() :
    goblinDataObject(Q.Context(),1)
{
    n = Q.n;
    next = Q.next;
    first = n;
    master = false;
    length = 0;

    LogEntry(LOG_MEM,"...Static queue instanciated");
}


template <class TItem,class TKey>
unsigned long staticQueue<TItem,TKey>::Size() throw()
{
    return
          sizeof(staticQueue<TItem,TKey>)
        + goblinDataObject::Allocated()
        + staticQueue::Allocated();
}


template <class TItem,class TKey>
unsigned long staticQueue<TItem,TKey>::Allocated() throw()
{
    unsigned long ret = 0;

    if (master) ret += n*sizeof(TItem);
    
    return ret;
}


template <class TItem,class TKey>
char* staticQueue<TItem,TKey>::Display() throw()
{
    LogEntry(MSG_TRACE,"Queue");
    
    if (Empty()) LogEntry(MSG_TRACE2,"    ---");
    else
    {
        TItem i = first;
        TItem counter = 0;
        
        THandle LH = LogStart(MSG_TRACE2,"   ");

        for (;i!=last;i=next[i])
        {
            if (counter>0 && (counter++)%10==0)
            {
                LogEnd(LH);
                LH = LogStart(MSG_TRACE2,"   ");
            }

            sprintf(CT.logBuffer,"%ld, ",i);
            LogAppend(LH,CT.logBuffer);
        }

        if (counter%10==0)
        {
            LogEnd(LH);
            LH = LogStart(MSG_TRACE2,"   ");
        }

        sprintf(CT.logBuffer,"%ld (last in)",i);
        LogEnd(LH,CT.logBuffer);
    }
    
    return NULL;
}


template <class TItem,class TKey>
staticQueue<TItem,TKey>::~staticQueue() throw()
{
    if (master) delete[] next;
    else
    {
        while (!Empty()) Delete();
    }

    LogEntry(LOG_MEM,"...Static queue disallocated");
}


template <class TItem,class TKey>
void staticQueue<TItem,TKey>::Insert(TItem w,TKey alpha) throw(ERRange,ERCheck)
{
    #if defined(_FAILSAVE_)

    if (w>=n) NoSuchItem("Insert",w);

    #endif
    
    if (next[w]!=n)
    {
        sprintf(CT.logBuffer,"%ld is already on the queue",w);
        Error(ERR_CHECK,"Insert",CT.logBuffer);
    }
    
    if (Empty()) last=first=w;
    else
    {
        next[last] = w;
        last = w;
    }

    next[last] = last;
    length++;
}


template <class TItem,class TKey>
TItem staticQueue<TItem,TKey>::Delete() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Delete","Queue is empty");

    #endif

    TItem temp = first;
    first = next[first];
    if (temp==first) first=n;
    next[temp] = n;
    length--;
    return temp;
}


template <class TItem,class TKey>
TItem staticQueue<TItem,TKey>::Peek() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Peek","Queue is empty");

    #endif
    
    return first;
}


template class staticQueue<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class staticQueue<TArc,TFloat>;

#endif
