
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "statstaq.h"
    

template <class TItem,class TKey>
staticStack<TItem,TKey>::staticStack(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    n = nn;
    prev = new TItem[n];
    top = n;
    master = true;
    depth = 0;
    
    for (TItem i=0;i<n;i++) prev[i] = n;

    LogEntry(LOG_MEM,"...Static stack instanciated");
};       


template <class TItem,class TKey>
staticStack<TItem,TKey>::staticStack(staticStack<TItem,TKey> &S) throw()
{
    n = S.n;
    prev = S.prev;
    top = n;
    master = false;
    depth = 0;
};       


template <class TItem,class TKey>
unsigned long staticStack<TItem,TKey>::Size() throw()
{
    return
          sizeof(staticStack<TItem,TKey>)
        + goblinDataObject::Allocated()
        + staticStack::Allocated();
};       


template <class TItem,class TKey>
unsigned long staticStack<TItem,TKey>::Allocated() throw()
{
    unsigned long ret = 0;

    if (master) ret += n*sizeof(TItem);
    
    return ret;
};       


template <class TItem,class TKey>
char* staticStack<TItem,TKey>::Display() throw()
{
    LogEntry(MSG_TRACE,"Stack");
    
    if (Empty()) LogEntry(MSG_TRACE2,"    ---");
    else
    {
        TItem i = top;
        TItem counter = 0;

        for (;prev[i]!=n;i=prev[i])
        {
            if ((counter++)%10==0) LogEntry(MSG_TRACE2,"   ");

            sprintf(CT.logBuffer,"%ld, ",i);
            LogEntry(MSG_APPEND,CT.logBuffer);
        }

        if (counter%10==0) LogEntry(MSG_TRACE2,"   ");

        sprintf(CT.logBuffer,"%ld (bottom)",i);
        LogEntry(MSG_APPEND,CT.logBuffer);
    }
    
    return NULL;
};       


template <class TItem,class TKey>
staticStack<TItem,TKey>::~staticStack() throw()
{
    delete[] prev;

    LogEntry(LOG_MEM,"...Static stack disallocated");
};


template <class TItem,class TKey>
void staticStack<TItem,TKey>::Insert(TItem w,TKey alpha) throw(ERRange,ERCheck)
{
    #if defined(_FAILSAVE_)

    if (w>=n) NoSuchItem("Insert",w);

    if (prev[w]!=n)
    {
        sprintf(CT.logBuffer,"%ld is already on the stack",w);
        Error(ERR_CHECK,"Insert",CT.logBuffer);
    }

    #endif
    
    prev[w] = top;
    top = w;
    depth++;
};


template <class TItem,class TKey>
TItem staticStack<TItem,TKey>::Delete() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Delete","Queue is empty");

    #endif

    TItem temp = top;
    top = prev[top];
    prev[temp] = n;
    depth--;
    return temp;
};


template <class TItem,class TKey>
TItem staticStack<TItem,TKey>::Peek() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (Empty()) Error(ERR_REJECTED,"Peek","Queue is empty");

    #endif
                                   
    return top;
};


template class staticStack<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class staticStack<TArc,TFloat>;

#endif
