#include <goblin.h>


static goblinController *CT;


void myLogEventHandler(msgType msg,TModule mod,THandle OH,char *logText)
{
    CT -> PlainLogEventHandler(msg,mod,OH,logText);
}


int main(int ParamCount,char *ParamStr[])
{
    CT = new goblinController();
    CT -> traceLevel = 1;

    unsigned l = strlen(ParamStr[ParamCount-1]);
    char* fileIn = (char *)malloc(l+5);
    strcat(strcpy(fileIn,ParamStr[ParamCount-1]),".gob");
    char* fileOut = (char *)malloc(l+5);
    strcat(strcpy(fileOut,ParamStr[ParamCount-1]),".rst");
    char* logFile = (char *)malloc(l+5);
    strcat(strcpy(logFile,ParamStr[ParamCount-1]),".log");

    cout << endl << "File access:" << endl;
    int pc = CT->FindParam(ParamCount,ParamStr,"-silent");
    if (pc==0)
    {
        CT->logStream       = new ofstream(logFile);
        CT->logEventHandler = &myLogEventHandler;
        cout << " Writing transscript to " << logFile << "..." << endl; 
    }

    char* type = "unknown";
    try
    {
        goblinImport F(fileIn,*CT);
        type = F.Scan();
    }
    catch (...) {};

    abstractMixedGraph* G = NULL;
    if (strcmp(type,"dense_digraph")==0) G = new denseDiGraph(fileIn,*CT);
    else
    {
        if (strcmp(type,"digraph")==0) G = new diGraph(fileIn,*CT);
        else
        {
            if (strcmp(type,"dense_graph")==0) G = new denseGraph(fileIn,*CT);
            else
            {
                if (strcmp(type,"graph")==0) G = new graph(fileIn,*CT);
                else
                {
                    cout << "...not a valid graph object: " << fileIn << endl << endl;
                    exit(1);
                }
            }
        }
    }
    
    CT -> traceStep = 1;
    pc = CT->FindParam(ParamCount,ParamStr,"-1");
    if (pc>0) CT->meth1Tree = 1;
    cout << " Graph read from " << fileIn << endl;
    
    cout << " Writing solution to "
        << fileOut << "..." << endl << endl;
    
    TNode root = NoNode;
    pc = CT->FindParam(ParamCount,ParamStr,"-r");
    if (pc>0)
    {
        root = atoi(ParamStr[pc+1]);
        cout << "Root node: " << root << endl;
    }
        
    CT -> methPQ = 1;

    CT -> arcLabels = 5;
    
    CT -> Configure(ParamCount,ParamStr);
    if (CT->traceLevel>1) CT->traceLevel = 1;
    
    cout << endl << "Method selection:" << endl;
    cout << " Min-Tree method (methMinTree): " << (int)CT->methMinTree << endl;
    cout << " Heap method (methPQ): " << (int)CT->methPQ << endl;
    
    if (CT->meth1Tree) cout << endl  << "Computing minimum 1-Tree...";
    else cout << endl << "Computing minimum spanning tree...";

    int connected = 1;
    try
    {
        if (root!=NoNode) G -> MinTree(root);
        else G -> MinTree();
    }
    catch (ERRejected) {connected = 0;};

    cout << endl;

    if (connected)
    {
        G -> ReleaseLabels();
        
        int sh = CT->FindParam(ParamCount,ParamStr,"-sh");
        if (sh)
        {
            goblinExport F(fileOut);
            if (root!=NoNode) G -> WritePredecessors(&F);
            else G -> WriteSubgraph(&F);
        }
        else G -> Write(fileOut);

        cout << "...Minimum spanning tree found." << endl;
    }
    else cout << "...Graph is disconnected." << endl;
    
    delete G;
    
    cout << endl << "Allocated " << goblinMaxSize << " bytes." << endl << endl;

    if (CT->logStream != &clog)
    {
        delete CT->logStream;
        CT->logStream = &clog;
    }

    delete CT;
    
    return 1-connected;
}
