
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, March 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "gosh.h"


const int TCL_UNHANDLED = TCL_OK + TCL_ERROR;

int Goblin_Mixed_Graph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    mixedGraph* G = (mixedGraph*)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        int ret = Goblin_Sparse_Cmd(G,interp,argc,argv);
        if (ret!=TCL_UNHANDLED) return ret;

        return Goblin_Generic_Graph_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Mixed_Graph (ClientData clientData) throw()
{
    delete (mixedGraph*)clientData;
    
    return TCL_OK;
}
    

int Goblin_Sparse_Graph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    graph* G = (graph*)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        int ret = Goblin_Sparse_Cmd(G,interp,argc,argv);
        if (ret!=TCL_UNHANDLED) return ret;

        return Goblin_Undirected_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Sparse_Graph (ClientData clientData) throw()
{
    delete (graph *)clientData;
    
    return TCL_OK;
}
    

int Goblin_Dense_Graph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    denseGraph *G = (denseGraph *)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        return Goblin_Undirected_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Dense_Graph (ClientData clientData) throw()
{
    delete (denseGraph *)clientData;
    
    return TCL_OK;
}
    

int Goblin_Sparse_Bigraph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    biGraph *G = (biGraph *)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        if (argc==4 && strcmp(argv[1],"node")==0 &&
            strcmp(argv[3],"swap")==0)
        {
            TNode u = atol(argv[2]);
            TNode v = G -> SwapNode(u);
            sprintf(interp->result,"%ld",v);
            return TCL_OK;
        }

        int ret = Goblin_Sparse_Cmd(G,interp,argc,argv);
        if (ret!=TCL_UNHANDLED) return ret;

        return Goblin_Bipartite_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Sparse_Bigraph (ClientData clientData) throw()
{
    delete (biGraph *)clientData;
    
    return TCL_OK;
}
    

int Goblin_Dense_Bigraph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    denseBiGraph *G = (denseBiGraph *)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        return Goblin_Bipartite_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Dense_Bigraph (ClientData clientData) throw()
{
    delete (denseBiGraph *)clientData;
    
    return TCL_OK;
}


int Goblin_Sparse_Digraph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    diGraph *G = (diGraph *)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        int ret = Goblin_Sparse_Cmd(G,interp,argc,argv);
        if (ret!=TCL_UNHANDLED) return ret;

        return Goblin_Directed_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Sparse_Digraph (ClientData clientData) throw()
{
    delete (diGraph *)clientData;
    
    return TCL_OK;
}
    

int Goblin_Dense_Digraph_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    denseDiGraph *G = (denseDiGraph *)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        return Goblin_Directed_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Dense_Digraph (ClientData clientData) throw()
{
    delete (denseDiGraph *)clientData;
    
    return TCL_OK;
}
    

int Goblin_Balanced_FNW_Cmd (ClientData clientData,Tcl_Interp* interp,
    int argc,_CONST_QUAL_ char* argv[])
{
    balancedFNW *G = (balancedFNW *)clientData;
    
    if (setjmp(goblinJumpBuffer) != 0) return TCL_ERROR;
    else try
    {
        if (strcmp(argv[1],"maxbalflow")==0)
        {
            if (argc!=3)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            TNode s = atol(argv[2]);
            G -> MaxBalFlow(s);
        
            try
            {
                sprintf(interp->result,"%f",G->FlowValue(s,s^1));
                return TCL_OK;
            }
            catch (ERCheck)
            {
                sprintf(interp->result,"Flow is corrupted");
                return TCL_ERROR;
            }
        }
    
        if (strcmp(argv[1],"mincbalflow")==0)
        {
            if (argc!=3)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            TNode s = atol(argv[2]);
            TFloat ret = G -> MinCBalFlow(s);

            try
            {
                G -> FlowValue(s,s^1);
                sprintf(interp->result,"%f",ret);
                return TCL_OK;
            }
            catch (ERCheck)
            {
                sprintf(interp->result,"Flow is corrupted");
                return TCL_ERROR;
            }
        }
    
        int ret = Goblin_Sparse_Cmd(G,interp,argc,argv);
        if (ret!=TCL_UNHANDLED) return ret;

        return Goblin_Directed_Cmd(G,interp,argc,argv);
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
}


int Goblin_Delete_Balanced_FNW (ClientData clientData) throw()
{
    delete (balancedFNW *)clientData;
    
    return TCL_OK;
}
    

int Goblin_Sparse_Cmd (abstractMixedGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc==2 && strcmp(argv[1],"planarize")==0)
    {
        if (G->Planarize())
            interp->result = "1";
        else interp->result = "0";
        return TCL_OK;
    }

    if (argc==3 && strcmp(argv[1],"exterior")==0)
    {
        TArc a = atol(argv[2]);
        G->MarkExteriorFace(a);
        interp->result = "";
        return TCL_OK;
    }

    if (argc==2 && strcmp(argv[1],"outerplanar")==0)
    {
        G->GrowExteriorFace();
        interp->result = "";
        return TCL_OK;
    }

    if (argc==2 && strcmp(argv[1],"convex")==0)
    {
        G->StraightLineDrawing();
        interp->result = "";
        return TCL_OK;
    }

    return TCL_UNHANDLED;
}
    

int Goblin_Bipartite_Cmd (abstractBiGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (strcmp(argv[1],"#outer")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->N1());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"#inner")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->N2());
        return TCL_OK;
    }

    return Goblin_Undirected_Cmd(G,interp,argc,argv);
}
    

int Goblin_Undirected_Cmd (abstractGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc<2)
    {
        interp->result = "Missing arguments";
        return TCL_ERROR;
    }

    if (strcmp(argv[1],"subgraph")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        graph *H = new graph(*G,
            goblinDataObject::OPT_SUB|goblinDataObject::OPT_PARALLELS);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"metric")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        denseGraph *H = new metricGraph(*G);

        Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Graph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"tiling")==0)
    {
        if (argc!=5)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        TNode x = atol(argv[3]);
        TNode y = atol(argv[4]);
        graph *H = new tiling(*G,(TOption)0,x,y);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"maxmatch")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TFloat ret = G -> MaximumMatching();
    
        sprintf(interp->result,"%f",ret);
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"mincmatch")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        if (G -> MinCMatching())
        {
            sprintf(interp->result,"%f",G->Weight());
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"tjoin")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        try
        {
            G -> MinCTJoin();
            sprintf(interp->result,"%f",G->Weight());
            return TCL_OK;
        }
        catch (ERRejected)
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    return Goblin_Generic_Graph_Cmd(G,interp,argc,argv);
}


int Goblin_Directed_Cmd (abstractDiGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc<2)
    {
        interp->result = "Missing arguments";
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"subgraph")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        diGraph *H = new diGraph(*G,
            goblinDataObject::OPT_SUB|goblinDataObject::OPT_PARALLELS);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Digraph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Digraph);

        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"splitgraph")==0)
    {
        if (argc!=5)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        TNode s = atol(argv[3]);
        TNode t = atol(argv[4]);
        balancedFNW *H = new splitGraph(*G,s,t);

        Tcl_CreateCommand(interp,argv[2],Goblin_Balanced_FNW_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Balanced_FNW);

        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"topsort")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode v = G->TopSort();

        if (v==NoNode)
        {
            interp->result = "*";
            return TCL_OK;
        }
        
        sprintf(interp->result,"%ld",v);
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"critical")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode v = G->CriticalPath();

        if (v==NoNode)
        {
            interp->result = "Graph is not a DAG";
            return TCL_ERROR;
        }
        
        sprintf(interp->result,"%ld",v);
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"maxflow")==0)
    {
        if (argc!=4)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode s = atol(argv[2]);
        TNode t = atol(argv[3]);
        G -> MaxFlow(s,t);
        
        try
        {
            sprintf(interp->result,"%f",G->FlowValue(s,t));
            return TCL_OK;
        }
        catch (ERCheck)
        {
            sprintf(interp->result,"Flow is corrupted");
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"mincflow")==0)
    {
        if (argc!=4)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode s = atol(argv[2]);
        TNode t = atol(argv[3]);
        TFloat ret = G -> MinCostFlow(s,t);

        try
        {
            G -> FlowValue(s,t);
            sprintf(interp->result,"%f",ret);
            return TCL_OK;
        }
        catch (ERCheck)
        {
            sprintf(interp->result,"Flow is corrupted");
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"circulation")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (G->AdmissibleCirculation())
        {
            interp->result = "";
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"minccirc")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (G->AdmissibleCirculation())
        {
            TFloat ret = G->MinCostCirculation();
            sprintf(interp->result,"%f",ret);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }

    return Goblin_Generic_Graph_Cmd(G,interp,argc,argv);
}


int Goblin_Generic_Graph_Cmd (abstractMixedGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc<2)
    {
        interp->result = "Missing arguments";
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"linegraph")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        graph *H = new lineGraph(*G,(TOption)0);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"dualgraph")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
        
        graph *H = new dualGraph(*G,(TOption)0);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"complement")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        graph *H = new complementaryGraph(*G,(TOption)0);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"underlying")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        graph *H = new graph(*G);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"induced")==0)
    {
        if (strcmp(argv[2],"subgraph")==0)
        {
            if (argc!=5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            TNode c = atol(argv[4]);
            mixedGraph *H = new inducedSubgraph(*G,c,TOption(0));

            Tcl_CreateCommand(interp,argv[3],Goblin_Mixed_Graph_Cmd,(ClientData)H,
                (Tcl_CmdDeleteProc *)Goblin_Delete_Mixed_Graph);

            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"orientation")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            diGraph *H = new inducedOrientation(*G,goblinDataObject::OPT_PARALLELS);

            Tcl_CreateCommand(interp,argv[3],Goblin_Sparse_Digraph_Cmd,(ClientData)H,
                (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Digraph);

            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"bigraph")==0)
        {
            if (argc!=6)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            TNode c1 = atol(argv[4]);
            TNode c2 = atol(argv[5]);
            biGraph *H = new inducedBigraph(*G,c1,c2,TOption(0));

            Tcl_CreateCommand(interp,argv[3],Goblin_Sparse_Bigraph_Cmd,(ClientData)H,
                (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Bigraph);

            interp->result = "";
            return TCL_OK;
        }
    }

    if (strcmp(argv[1],"contraction")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        mixedGraph *H = new colourContraction(*G,TOption(0));

        Tcl_CreateCommand(interp,argv[2],Goblin_Mixed_Graph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Mixed_Graph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"nodesplitting")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        diGraph *H = new nodeSplitting(*G,0);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Digraph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Digraph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"orientation")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        diGraph *H = new diGraph(*G,goblinDataObject::OPT_PARALLELS);

        Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Digraph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Digraph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"distances")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        denseDiGraph *H = new distanceGraph(*G);

        Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Digraph_Cmd,(ClientData)H,
            (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Digraph);

        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"merge")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
       
        goblinAbstractObject *X =
            (G->Context()).ObjectPointer(THandle(atol(argv[2])));

        if (!X->IsGraphObject())
        {
            sprintf(interp->result,"Not a graph object ID: %s",argv[3]);
            return TCL_ERROR;
        }
        
        abstractMixedGraph* Y = NULL;
        
        if (X->ClassID()==IDMixedGraph)
        {
            mixedGraph* Z = (mixedGraph*)X->BackCast();
            Y = Z;
        }

        if (X->ClassID()==IDSparseGraph)
        {
            graph* Z = (graph*)X->BackCast();
            Y = Z;
        }

        if (X->ClassID()==IDSparseDiGraph)
        {
            diGraph* Z = (diGraph*)X->BackCast();
            Y = Z;
        }

        if (X->ClassID()==IDSparseBiGraph)
        {
            biGraph* Z = (biGraph*)X->BackCast();
            Y = Z;
        }

        if (Y==NULL)
        {
            interp->result = "Unhandled object type";
            return TCL_ERROR;
        }
        else
        {
            G -> Merge(*Y);

            interp->result = "";
            return TCL_OK;
        }
    }

    if (strcmp(argv[1],"layout")==0)
    {
        if (argc<3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"scale")==0)
        {
            if (argc!=5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            TFloat dx = atof(argv[3]);
            TFloat dy = atof(argv[4]);
            G -> ScaleEmbedding(dx,dy);
        
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"tree")==0)
        {
            if (argc!=3 && argc!=5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            try
            {
                if (argc==5)
                {
                    TFloat dx = atof(argv[3]);
                    TFloat dy = atof(argv[4]);
                    G -> TreeLayout(dx,dy);
                }
                else G -> TreeLayout();
            }
            catch (ERRejected) {}
        
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"spring")==0)
        {
            if (argc!=3 && argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            try
            {
                if (argc==4)
                {
                    TFloat gamma = atof(argv[3]);
                    G -> SpringEmbedding(gamma);
                }
                else G -> SpringEmbedding();
            }
            catch (ERRejected) {}
        
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"colours")==0)
        {
            if (argc!=3 && argc!=5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            if (argc==5)
            {
                TFloat dx = atof(argv[3]);
                TFloat dy = atof(argv[4]);
                G -> ColourLayout(dx,dy);
            }
            else G -> ColourLayout();
        
            interp->result = "";
            return TCL_OK;
        };
   
        if (strcmp(argv[2],"circle")==0)
        {
            if (argc!=3 && argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            if (argc==4)
            {
                TFloat sep = atof(argv[3]);
                G -> CircleLayout(sep);
            }
            else G -> CircleLayout();
        
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"grid")==0)
        {
            if (argc!=3 && argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            if (argc==4)
            {
                TFloat sep = atof(argv[3]);
                G -> GridLayout(sep);
            }
            else G -> GridLayout();
        
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"align")==0)
        {
            if (argc!=3 && argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            if (argc==4)
            {
                TFloat sep = atof(argv[3]);
                G -> AutoArcAlignment(sep);
            }
            else G -> AutoArcAlignment();
            
            interp->result = "";
            return TCL_OK;
        }
    
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        if (strcmp(argv[2],"strip")==0)
        {
            G -> StripEmbedding();
            interp->result = "";
            return TCL_OK;
        }
    
        sprintf(interp->result,"Unknown option: %s layout %s",argv[0],argv[2]);
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"extract")==0)
    {
        if (argc<3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"tree")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            try
            {
                TNode r = atol(argv[3]);
                G -> ExtractTree(r);
                interp->result = "1";
            }
            catch (ERCheck)
            {
                interp->result = "0";
            }
        
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"path")==0)
        {
            if (argc!=5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            try
            {
                TNode s = atol(argv[3]);
                TNode t = atol(argv[3]);
                G -> ExtractPath(s,t);
                interp->result = "1";
            }
            catch (ERCheck)
            {
                interp->result = "0";
            }
        
            return TCL_OK;
        }
    
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"cycles")==0)
        {
            try
            {
                sprintf(interp->result,"%ld",G -> ExtractCycles());
            }
            catch (ERCheck)
            {
                interp->result = "0";
            }
        
            return TCL_OK;
        }
    
        try
        {
            interp->result = "1";

            if (strcmp(argv[2],"matching")==0)
            {
                G -> Extract1Matching();
                return TCL_OK;
            }

            if (strcmp(argv[2],"trees")==0)
            {
                G -> ExtractTrees();
                return TCL_OK;
            }

            if (strcmp(argv[2],"cut")==0)
            {
                G -> ExtractCut();
                return TCL_OK;
            }

            if (strcmp(argv[2],"bipartition")==0)
            {
                G -> ExtractBipartition();
                return TCL_OK;
            }

            if (strcmp(argv[2],"colours")==0)
            {
                G -> ExtractColours();
                return TCL_OK;
            }


            interp->result = "1";
        }
        catch (ERCheck)
        {
            interp->result = "0";
            return TCL_OK;
        }
        
        sprintf(interp->result,"Unknown option: %s extract %s",argv[0],argv[2]);
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"delete")==0)
    {
        if (argc==2)
        {
            Tcl_DeleteCommand(interp,argv[0]);
            interp->result = "";
            return TCL_OK;
        }

        if (argc==3)
        {
            if (strcmp(argv[2],"subgraph")==0)
            {
                G -> InitSubgraph();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"labels")==0)
            {
                G -> ReleaseLabels();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"predecessors")==0)
            {
                G -> ReleasePredecessors();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"colours")==0)
            {
                G -> ReleaseColours();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"potentials")==0)
            {
                G -> ReleasePotentials();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"partition")==0)
            {
                G -> ReleasePartition();
                interp->result = "";
                return TCL_OK;
            }

            sprintf(interp->result,"Unknown option: %s delete %s",argv[0],argv[2]);
            return TCL_ERROR;
        }

        interp->result = "Wrong number of arguments";
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"#nodes")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->N());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"#artificial")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->NI());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"#arcs")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->M());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"source")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        if (G->Source()==NoNode) interp->result = "*";
        else sprintf(interp->result,"%ld",G->Source());

        return TCL_OK;
    }
    
    if (strcmp(argv[1],"target")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        if (G->Target()==NoNode) interp->result = "*";
        else sprintf(interp->result,"%ld",G->Target());

        return TCL_OK;
    }
    
    if (strcmp(argv[1],"root")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        if (G->Root()==NoNode) interp->result = "*";
        else sprintf(interp->result,"%ld",G->Root());

        return TCL_OK;
    }
    
    if (strcmp(argv[1],"cardinality")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%f",G->Cardinality());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"weight")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%f",G->Weight());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"length")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%f",G->Length());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"generate")==0)
    {
        if (strcmp(argv[2],"length")==0 && argc==3)
        {
            for (TArc a=0;a<G->M();a++)
                G->SetLength(2*a,TFloat(G->Context().SignedRand()));

            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"ucap")==0 && argc==3)
        {
            for (TArc a=0;a<G->M();a++)
                G->SetUCap(2*a,G->LCap(2*a)+1+TCap(G->Context().UnsignedRand()));

            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"lcap")==0 && argc==3)
        {
            for (TArc a=0;a<G->M();a++)
            {
                TCap oldCap = G->LCap(2*a);
                TCap newCap = TCap(TCap(G->Context().UnsignedRand()));

                if (newCap>oldCap)
                {
                    G -> SetUCap(2*a,G->UCap(2*a)+newCap-oldCap);
            	    G -> SetLCap(2*a,newCap);
                }
                else
                {
            	    G -> SetLCap(2*a,newCap);
                    G -> SetUCap(2*a,G->UCap(2*a)+newCap-oldCap);
                }
            }

            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"geometry")==0 && argc==3)
        {
            for (TNode v=0;v<G->N();v++)
            {
                G->SetC(v,0,TFloat(G->Context().SignedRand()));
                G->SetC(v,1,TFloat(G->Context().SignedRand()));
            }

            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"parallels")==0)
        {
            G -> ExplicitParallels();
            interp->result = "";
            return TCL_OK;
        }
    
        if (argc!=4)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        if (strcmp(argv[2],"arcs")==0)
        {
            G -> RandomArcs(TArc(atol(argv[3])));
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"eulerian")==0)
        {
            G -> RandEulerian(TArc(atol(argv[3])));
            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"regular")==0)
        {
            G -> RandRegular(TNode(atol(argv[3])));
            interp->result = "";
            return TCL_OK;
        }

        sprintf(interp->result,"Unknown option: %s generate %s",argv[0],argv[2]);
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"write")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        G -> Write(argv[2]);
        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"spath")==0)
    {
        if (argc>4 || argc<3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode s = atol(argv[2]);
        TNode t = NoNode;

        if (argc==4 && strcmp(argv[3],"*")!=0) t=atol(argv[3]);

        bool solved = false;
        try 
        {
            G -> ShortestPath(s,t);
            solved = true;
        }
        catch (...) {}
        
        G -> InitPotentials();
        G -> UpdatePotentials(InfFloat);

        if (solved && (t==NoNode || G->Dist(t)<InfFloat))
        {
            sprintf(interp->result,"%f",G->Length());
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"connected")==0)
    {
        if (argc==2)
        {
            sprintf(interp->result,"%d",G->Connected());
            return TCL_OK;
        }
        
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode k = atol(argv[2]);
        sprintf(interp->result,"%d",G->Connected(k));
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"sconnected")==0)
    {
        if (argc==2)
        {
            sprintf(interp->result,"%d",G->StronglyConnected());
            return TCL_OK;
        }
        
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode k = atol(argv[2]);
        sprintf(interp->result,"%d",G->StronglyConnected(k));
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"econnected")==0)
    {
        if (argc==2)
        {
            sprintf(interp->result,"%d",G->Connected());
            return TCL_OK;
        }
        
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode k = atol(argv[2]);
        sprintf(interp->result,"%d",G->EdgeConnected(k));
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"seconnected")==0)
    {
        if (argc==2)
        {
            sprintf(interp->result,"%d",G->StronglyConnected());
            return TCL_OK;
        }
        
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        TNode k = atol(argv[2]);
        sprintf(interp->result,"%d",G->StronglyEdgeConnected(k));
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"stnumbering")==0)
    {
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
        
        TArc returnArc = NoArc;
        
        if (argc==3) returnArc = atol(argv[2]);

        sprintf(interp->result,"%d",G->STNumbering(returnArc));
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"colouring")==0)
    {
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode chi = 0;

        if (argc==3) chi = G->NodeColouring(atol(argv[2]));
        else chi = G->NodeColouring();

        if (chi>0)
        {
            sprintf(interp->result,"%ld",chi);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"edgecolouring")==0)
    {
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode chi = 0;

        if (argc==3) chi = G->EdgeColouring(atol(argv[2]));
        else chi = G->EdgeColouring();

        if (chi>0)
        {
            sprintf(interp->result,"%ld",chi);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"cliques")==0)
    {
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode chi = 0;

        if (argc==3) chi = G->CliqueCover(atol(argv[2]));
        else chi = G->CliqueCover();

        if (chi>0)
        {
            sprintf(interp->result,"%ld",chi);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"stable")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->StableSet());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"clique")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->Clique());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"vertexcover")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->VertexCover());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"eulerian")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%d",G->EulerCycle());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"tsp")==0)
    {
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode r = NoNode;

        if (argc==3 && strcmp(argv[2],"*")!=0) r=atol(argv[2]);

        TFloat ret = G->TSP(r);

        if (ret!=InfFloat)
        {
            sprintf(interp->result,"%f",ret);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"mintree")==0 || strcmp(argv[1],"min1tree")==0)
    {
        if (argc>3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode v = NoNode;

        if (argc==3 && strcmp(argv[2],"*")!=0) v = atol(argv[2]);

        TFloat ret = InfFloat;

        if (strcmp(argv[1],"mintree")==0)
            ret = G->MinTree(v);
        else ret = G->MinTree(v,abstractMixedGraph::MTREE_1TREE);

        if (ret!=InfFloat)
        {
            sprintf(interp->result,"%f",ret);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"steiner")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode v = atol(argv[2]);

        if (strcmp(argv[2],"*")==0) v = NoNode;

        TFloat ret = G->SteinerTree(v);

        if (ret!=InfFloat)
        {
            sprintf(interp->result,"%f",ret);
            return TCL_OK;
        }
        else
        {
            interp->result = "No such structure exists";
            return TCL_ERROR;
        }
    }
    
    if (strcmp(argv[1],"postman")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        G -> ChinesePostman();
        sprintf(interp->result,"%f",G->Weight());
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"treepacking")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode v = NoNode;
        if (argc==3) v = atol(argv[2]);

        TCap ret = G->TreePacking(v);
        sprintf(interp->result,"%f",ret);
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"maxcut")==0)
    {
        if (argc!=2 && argc!=4)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        TNode s = NoNode;
        TNode t = NoNode;

        if (argc==4)
        {
            if (strcmp(argv[2],"*")==0) s = NoNode;
            else s = atol(argv[2]);
            
            if (strcmp(argv[3],"*")==0) t = NoNode;
            else t = atol(argv[3]);
        }

        sprintf(interp->result,"%f",G->MaxCut(s,t));
	
		G -> InitSubgraph();

		for (TArc a=0;a<G->M();a++)
        {
		    if (G->Colour(G->StartNode(2*a))!=G->Colour(G->EndNode(2*a)) &&
                (!G->Blocking(2*a+1) || G->Colour(G->StartNode(2*a))==0))
			    G -> AddArc(2*a,G->UCap(2*a)-G->Sub(2*a));
		    else G -> OmitArc(2*a,G->Sub(2*a)-G->LCap(2*a));
        }
	
        return TCL_OK;
    }
    
    if (strcmp(argv[1],"max")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"ucap")==0)
        {
            if (G->MaxUCap()!=InfCap)
            {
                if (G->MaxUCap()!=(long int)(G->MaxUCap()))
                    sprintf(interp->result,"%f",G->MaxUCap());
                else  sprintf(interp->result,"%ld",(long int)G->MaxUCap());
            }
            else interp->result = "*";

            return TCL_OK;
        }

        if (strcmp(argv[2],"lcap")==0)
        {
            if (G->MaxLCap()!=(long int)(G->MaxLCap()))
                sprintf(interp->result,"%f",G->MaxLCap());
            else  sprintf(interp->result,"%ld",(long int)G->MaxLCap());

            return TCL_OK;
        }

        if (strcmp(argv[2],"length")==0)
        {
            if (G->MaxLength()!=InfFloat)
            {
                if (G->MaxLength()!=(long int)(G->MaxLength()))
                    sprintf(interp->result,"%f",G->MaxLength());
                else  sprintf(interp->result,"%ld",(long int)G->MaxLength());
            }
            else interp->result = "*";

            return TCL_OK;
        }

        if (strcmp(argv[2],"demand")==0)
        {
            if (G->MaxDemand()!=InfCap)
            {
                if (G->MaxDemand()!=(long int)(G->MaxDemand()))
                    sprintf(interp->result,"%f",G->MaxDemand());
                else  sprintf(interp->result,"%ld",(long int)G->MaxDemand());
            }
            else interp->result = "*";

            return TCL_OK;
        }

        if (strcmp(argv[2],"cx")==0)
        {
            if (G->CMax(0)!=(long int)(G->CMax(0)))
                sprintf(interp->result,"%f",G->CMax(0));
            else  sprintf(interp->result,"%ld",(long int)G->CMax(0));

            return TCL_OK;
        }

        if (strcmp(argv[2],"cy")==0)
        {
            if (G->CMax(1)!=(long int)(G->CMax(1)))
                sprintf(interp->result,"%f",G->CMax(1));
            else  sprintf(interp->result,"%ld",(long int)G->CMax(1));

            return TCL_OK;
        }
    
        sprintf(interp->result,"Unknown option: %s max %s",argv[0],argv[2]);
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"constant")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"ucap")==0)
        {
            sprintf(interp->result,"%d",G->CUCap());
            return TCL_OK;
        }

        if (strcmp(argv[2],"lcap")==0)
        {
            sprintf(interp->result,"%d",G->CLCap());
            return TCL_OK;
        }

        if (strcmp(argv[2],"length")==0)
        {
            sprintf(interp->result,"%d",G->CLength());
            return TCL_OK;
        }

        if (strcmp(argv[2],"demand")==0)
        {
            sprintf(interp->result,"%d",G->CDemand());
            return TCL_OK;
        }
    
        sprintf(interp->result,"Unknown option: %s constant %s",argv[0],argv[2]);
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"set")==0)
    {
        if (argc!=4)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"ucap")==0)
        {
            TCap uu = InfCap;

            if (strcmp(argv[3],"*")!=0) uu = TCap(atof(argv[3]));

            G -> SetCUCap(uu);
            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"lcap")==0)
        {
            TCap ll = TCap(atof(argv[3]));
            G -> SetCLCap(ll);
            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"length")==0)
        {
            TFloat ll = InfFloat;

            if (strcmp(argv[3],"*")!=0) ll = TFloat(atof(argv[3]));

            G -> SetCLength(ll);
            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"demand")==0)
        {
            TCap bb = TCap(atof(argv[3]));
            G -> SetCDemand(bb);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[2],"source")==0)
        {
            TNode v = NoNode;

            if (strcmp(argv[3],"*")!=0) v = TNode(atol(argv[3]));

            G -> SetSourceNode(v);
            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"target")==0)
        {
            TNode v = NoNode;

            if (strcmp(argv[3],"*")!=0) v = TNode(atol(argv[3]));

            G -> SetTargetNode(v);
            interp->result = "";
            return TCL_OK;
        }

        if (strcmp(argv[2],"root")==0)
        {
            TNode v = NoNode;

            if (strcmp(argv[3],"*")!=0) v = TNode(atol(argv[3]));

            G -> SetRootNode(v);
            interp->result = "";
            return TCL_OK;
        }
    }
    
    if (strcmp(argv[1],"is")==0)
    {
        if (strcmp(argv[2],"graph")==0)
        {
            interp->result = "1";
            return TCL_OK;
        }

        if (strcmp(argv[2],"sparse")==0)
        {
            if (G->IsSparse()) interp->result = "1";
            else interp->result = "0";

            return TCL_OK;
        }

        if (strcmp(argv[2],"directed")==0)
        {
            if (G->IsDirected()) interp->result = "1";
            else interp->result = "0";

            return TCL_OK;
        }

        if (strcmp(argv[2],"undirected")==0)
        {
            if (G->IsUndirected()) interp->result = "1";
            else interp->result = "0";

            return TCL_OK;
        }
    
        if (strcmp(argv[2],"bipartite")==0)
        {
            if (G->IsBipartite()) interp->result = "1";
            else interp->result = "0";

            return TCL_OK;
        }
    
        if (strcmp(argv[2],"balanced")==0)
        {
            if (G->IsBalanced()) interp->result = "1";
            else interp->result = "0";

            return TCL_OK;
        }

        if (strcmp(argv[2],"planar")==0)
        {
            if (G->IsPlanar()) interp->result = "1";
            else interp->result = "0";

            return TCL_OK;
        }

        // Do not complain, message 'is' is interpreted again
    }
    
    if (strcmp(argv[1],"adjacency")==0)
    {
        if (argc!=4)
        {
            interp->result = "Missing end nodes";
            return TCL_ERROR;
        }
    
        TNode u = (TArc)atol(argv[2]);
        TNode v = (TArc)atol(argv[3]);
        TArc a = G->Adjacency(u,v);
        
        if (a==NoArc) interp->result = "*";
        else sprintf(interp->result,"%ld",a);

        return TCL_OK;
    }
    
    if (strcmp(argv[1],"node")==0)
        return Goblin_Node_Cmd(G,interp,argc,argv);
    
    if (strcmp(argv[1],"arc")==0)
        return Goblin_Arc_Cmd(G,interp,argc,argv);
    
    return Goblin_Generic_Cmd(G,interp,argc,argv);
}
    

int Goblin_Node_Cmd (abstractMixedGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc<3)
    {
        interp->result = "Wrong number of arguments";
        return TCL_ERROR;
    }
    
    if (strcmp(argv[2],"insert")==0)
    {
        if (argc!=3)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",G->InsertNode());
        return TCL_OK;
    }
    
    if (argc<4)
    {
        interp->result = "Wrong number of arguments";
        return TCL_ERROR;
    }
    
    TNode v = atol(argv[2]);
    
    if (strcmp(argv[3],"delete")==0)
    {
        G->DeleteNode(v);
        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"demand")==0)
    {
        sprintf(interp->result,"%g",G->Demand(v));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"first")==0)
    {
        TArc a = G->First(v);
        if (a==NoArc) interp->result = "*";
        else sprintf(interp->result,"%ld",a);
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"cx")==0)
    {
        sprintf(interp->result,"%f",G->C(v,0));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"cy")==0)
    {
        sprintf(interp->result,"%f",G->C(v,1));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"hidden")==0)
    {
        sprintf(interp->result,"%d",G->HiddenNode(v));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"colour")==0)
    {
        if (G->Colour(v)!=NoNode)
            sprintf(interp->result,"%ld",G->Colour(v));
        else interp->result = "*";

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"degree")==0)
    {
        sprintf(interp->result,"%g",G->Deg(v));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"distance")==0)
    {
        if (G->Dist(v)!=InfFloat)
            sprintf(interp->result,"%g",G->Dist(v));
        else interp->result = "*";

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"potential")==0)
    {
        sprintf(interp->result,"%g",G->Pi(v));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"predecessor")==0)
    {
        if (G->Pred(v)!=NoArc)
            sprintf(interp->result,"%ld",G->Pred(v));
        else interp->result = "*";

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"first")==0)
    {
        sprintf(interp->result,"%ld",G->First(v));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"thread")==0)
    {
        if (G->Interpolate(v)!=NoNode)
            sprintf(interp->result,"%ld",G->Interpolate(v));
        else interp->result = "*";

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"set")==0)
    {
        if (strcmp(argv[4],"thread")==0)
        {
            if (argc<7 || !(argc&1))
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
        
            for (int i=5;i<argc;i+=2)
            {
                TFloat xx = TFloat(atof(argv[i]));
                TFloat yy = TFloat(atof(argv[i+1]));

                TNode w = G->InsertBendNode(v);
                G -> SetC(w,0,xx);
                G -> SetC(w,1,yy);

                v = w;
            }

            interp->result = "";
            return TCL_OK;
        }
    
        if (argc!=6)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[4],"demand")==0)
        {
            TCap bb = TCap(atof(argv[5]));
            G -> SetDemand(v,bb);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"cx")==0)
        {
            TFloat xx = TFloat(atof(argv[5]));
            G -> SetC(v,0,xx);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"cy")==0)
        {
            TFloat yy = TFloat(atof(argv[5]));
            G -> SetC(v,1,yy);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"first")==0)
        {
            TArc a = TArc(atol(argv[5]));
            G -> SetFirst(v,a);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"distance")==0)
        {
            TFloat dd = InfFloat;

            if (strcmp(argv[5],"*")!=0) dd = TFloat(atof(argv[5]));

            G -> SetDist(v,dd);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"potential")==0)
        {
            TFloat pp = TFloat(atof(argv[5]));
            G -> SetPotential(v,pp);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"colour")==0)
        {
            TNode cc = NoNode;

            if (strcmp(argv[5],"*")!=0) cc =  TNode(atoi(argv[5]));

            G -> SetColour(v,cc);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"predecessor")==0)
        {
            TArc aa = NoArc;

            if (strcmp(argv[5],"*")!=0) aa = TArc(atoi(argv[5]));

            G -> SetPred(v,aa);
            interp->result = "";
            return TCL_OK;
        }

        sprintf(interp->result,"Unknown option: %s node %s set %s",argv[0],argv[2],argv[4]);
        return TCL_ERROR;
    }
    
    sprintf(interp->result,"Unknown option: %s node %s",argv[0],argv[2]);
    return TCL_ERROR;
}
    

int Goblin_Arc_Cmd (abstractMixedGraph *G,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc<4)
    {
        interp->result = "Wrong number of arguments";
        return TCL_ERROR;
    }
    
    if (strcmp(argv[2],"insert")==0)
    {
        if (argc!=5)
        {
            interp->result = "Missing end nodes";
            return TCL_ERROR;
        }
    
        TNode u = TArc(atol(argv[3]));
        TNode v = TArc(atol(argv[4]));
        
        sprintf(interp->result,"%ld",G->InsertArc(u,v));
        return TCL_OK;
    }
    
    TArc a = (TArc)atol(argv[2]);
    
    if (strcmp(argv[3],"delete")==0)
    {
        G -> DeleteArc(a);
        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"straight")==0)
    {
        G -> ReleaseBendNodes(a);
        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[3],"hidden")==0)
    {
        sprintf(interp->result,"%d",G->HiddenArc(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"ucap")==0)
    {
        if (G->UCap(a)==InfCap) sprintf(interp->result,"*");
        else sprintf(interp->result,"%g",G->UCap(a));

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"lcap")==0)
    {
        sprintf(interp->result,"%g",G->LCap(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"length")==0)
    {
        sprintf(interp->result,"%g",G->Length(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"subgraph")==0)
    {
        sprintf(interp->result,"%g",G->Sub(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"colour")==0)
    {
        if (G->EdgeColour(a)!=NoArc)
            sprintf(interp->result,"%ld",G->EdgeColour(a));
        else interp->result = "*";

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"orientation")==0)
    {
        sprintf(interp->result,"%d",G->Orientation(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"head")==0)
    {
        sprintf(interp->result,"%ld",G->EndNode(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"tail")==0)
    {
        sprintf(interp->result,"%ld",G->StartNode(a));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"right")==0)
    {
        sprintf(interp->result,"%ld",G->Right(a,G->StartNode(a)));
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"align")==0)
    {
        if (G->Align(a)!=NoNode)
            sprintf(interp->result,"%ld",G->Align(a));
        else interp->result = "*";

        return TCL_OK;
    }
    
    if (strcmp(argv[3],"flip")==0)
    {
        G -> FlipArc(a);
        interp->result = "";
        return TCL_OK;
    }
    
    if (strcmp(argv[3],"set")==0)
    {
        if (strcmp(argv[4],"align")==0)
        {
    
            TFloat xx = TFloat(atof(argv[5]));
            TFloat yy = TFloat(atof(argv[6]));

            TNode v = G->Align(a);
            if (v==NoNode) v = G->InsertAlignmentPoint(a);

            G -> SetC(v,0,xx);
            G -> SetC(v,1,yy);

            sprintf(interp->result,"%ld",v);
            return TCL_OK;
        }
    
        if (argc!=6)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[4],"ucap")==0)
        {
            TCap uu = InfCap;

            if (strcmp(argv[5],"*")!=0) uu = TCap(atof(argv[5]));

            G -> SetUCap(a,uu);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"lcap")==0)
        {
            TCap ll = TCap(atof(argv[5]));
            G -> SetLCap(a,ll);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"length")==0)
        {
            TFloat ll = InfFloat;

            if (strcmp(argv[5],"*")!=0) ll = TFloat(atof(argv[5]));

            G -> SetLength(a,ll);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"right")==0)
        {
            TArc a2 = TArc(atol(argv[5]));
            G -> SetRight(a,a2);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"colour")==0)
        {
            TArc cc = NoArc;

            if (strcmp(argv[5],"*")!=0) cc =  TArc(atoi(argv[5]));
 
            G -> SetEdgeColour(a,cc);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"orientation")==0)
        {
            char oo = char(atol(argv[5]));
            G -> SetOrientation(a,oo);
            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[4],"subgraph")==0)
        {
            TFloat ll = (TFloat)atof(argv[5]);
            TFloat delta = ll-G->Sub(a);

            if (delta<0) G->OmitArc(a,-delta);
            if (delta>0) G->AddArc(a,delta);

            interp->result = "";
            return TCL_OK;
        }
    
        sprintf(interp->result,"Unknown option: %s arc %s set %s",argv[0],argv[2],argv[4]);
        return TCL_ERROR;
    }
    
    sprintf(interp->result,"Unknown option: %s arc %s",argv[0],argv[2]);
    return TCL_ERROR;
}
