
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, March 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "gosh.h"
#include "messenger.h"


goblinController *CT = NULL;
jmp_buf goblinJumpBuffer;

static goblinMessenger *MSG = NULL;
static long unsigned goblinInterprCounter = 0;
static char* transscriptName = NULL;
static Tcl_Interp *masterInterp = NULL;
static char returnCode[255] = "";
static int errorCode = TCL_OK;
    

void Goblin_Config() throw()
{
    char* configName = new char[strlen(getenv("HOME"))+9];
    sprintf(configName,"%s/.goshrc",getenv("HOME"));
    ifstream test(configName);

    if (test)
    {
        try
        {   
            goblinImport* F = new goblinImport(configName,*CT);
            F -> ReadConfiguration();
            delete F;

            F = new goblinImport(configName,goblinDefaultContext);
            F -> ReadConfiguration();
            delete F;

            sprintf(CT->logBuffer,
                 "...configuration file \"%s\" read",configName);
            CT -> LogEntry(LOG_IO,NoHandle,CT->logBuffer);
        }
        catch (...) {}
    }

    delete[] configName;
}

    
void Goblin_TerminateHandler()
{
    sprintf(CT->logBuffer,"...Thread terminated abnormally");
    CT -> LogEntry(LOG_SHELL,NoHandle,CT->logBuffer);

    sprintf(returnCode,"...Thread terminated abnormally");
    errorCode = TCL_ERROR;

    longjmp(goblinJumpBuffer,1);
}


int Goblin_Init(Tcl_Interp *interp)
{
    set_terminate(Goblin_TerminateHandler);

    if (goblinInterprCounter==0)
    {
        transscriptName = new char[strlen(getenv("HOME"))+10];
        sprintf(transscriptName,"%s/gosh.log",getenv("HOME"));
        masterInterp = interp;
        CT = new goblinController(goblinDefaultContext);
        CT->logStream         = new ofstream(transscriptName,ios::app);
        CT->traceLevel        = 0;
        CT->traceEventHandler = &Goblin_TraceEventHandler;
        CT->logEventHandler   = &Goblin_LogEventHandler;
        CT->stopSignalHandler = &Goblin_StopSignalHandler;
        MSG = new goblinMessenger(100,*CT);
        Goblin_Config();
    }
    else CT -> LogEntry(LOG_SHELL,NoHandle,"Starting GOSH interpreter...");

    goblinInterprCounter++;

    Tcl_CreateCommand(interp,"goblin",Goblin_Cmd,(ClientData)NULL,
                      (Tcl_CmdDeleteProc *)Goblin_Delete);
                      
    Tcl_SetVar(interp,"goblinMasterEvent","0",0);
                      
    Tcl_LinkVar(interp,"goblinXZoom",(char*)&(CT->xZoom),TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp,"goblinYZoom",(char*)&(CT->yZoom),TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp,"goblinXShift",(char*)&(CT->xShift),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinYShift",(char*)&(CT->yShift),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinNodeSize",(char*)&(CT->nodeSize),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinNodeSep",(char*)&(CT->nodeSep),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinArrowSize",(char*)&(CT->arrowSize),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLegenda",(char*)&(CT->legenda),TCL_LINK_INT);
                      
    Tcl_LinkVar(interp,"goblinArcLabels1",(char*)&(CT->arcLabels),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinArcLabels2",(char*)&(CT->arcLabels2),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinArcLabels3",(char*)&(CT->arcLabels3),TCL_LINK_INT);

    Tcl_LinkVar(interp,"goblinNodeFormatting",(char*)&(CT->nodeFormatting),
        TCL_LINK_STRING | TCL_LINK_READ_ONLY);
    Tcl_LinkVar(interp,"goblinArcFormatting",(char*)&(CT->arcFormatting),
        TCL_LINK_STRING | TCL_LINK_READ_ONLY);
    Tcl_LinkVar(interp,"goblinWallpaper",(char*)&(CT->wallpaper),
        TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    Tcl_LinkVar(interp,"goblinNodeLabels",(char*)&(CT->nodeLabels),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinNodeColours",(char*)&(CT->nodeColours),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinSubgraph",(char*)&(CT->subgraph),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinPredecessors",(char*)&(CT->predecessors),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinArcStyle",(char*)&(CT->arcStyle),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinNodeStyle",(char*)&(CT->nodeStyle),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinArrows",(char*)&(CT->arrows),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinDisplayMode",(char*)&(CT->displayMode),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinTraceLevel",(char*)&(CT->traceLevel),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinTraceData",(char*)&(CT->traceData),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinTraceStep",(char*)&(CT->traceStep),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinThreshold",(char*)&(CT->threshold),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinTraceCounter",(char*)&(CT->fileCounter),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinCommLevel",(char*)&(CT->commLevel),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinBreakLevel",(char*)&(CT->breakLevel),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogLevel",(char*)&(CT->logLevel),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinLogMethods",(char*)&(CT->logMeth),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogMemory",(char*)&(CT->logMem),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogManipulate",(char*)&(CT->logMan),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogIO",(char*)&(CT->logIO),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogResults",(char*)&(CT->logRes),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogWarnings",(char*)&(CT->logWarn),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogTimers",(char*)&(CT->logTimers),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinLogGaps",(char*)&(CT->logGaps),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinMethFailSave",(char*)&(CT->methFailSave),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethGeometry",(char*)&(CT->methGeometry),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethAdjacency",(char*)&(CT->methAdjacency),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethDSU",(char*)&(CT->methDSU),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethPQ",(char*)&(CT->methPQ),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethModLength",(char*)&(CT->methModLength),TCL_LINK_INT);

    Tcl_LinkVar(interp,"goblinMethSearch",(char*)&(CT->methSearch),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMinTree",(char*)&(CT->methMinTree),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMaxFlow",(char*)&(CT->methMaxFlow),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMinCut",(char*)&(CT->methMinCut),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMinCFlow",(char*)&(CT->methMinCFlow),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMinCCirc",(char*)&(CT->methMinCCirc),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethNWPricing",(char*)&(CT->methNWPricing),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMaxBalFlow",(char*)&(CT->methMaxBalFlow),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethBNS",(char*)&(CT->methBNS),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMinCBalFlow",(char*)&(CT->methMinCBalFlow),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethPrimalDual",(char*)&(CT->methPrimalDual),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethTSP",(char*)&(CT->methTSP),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethCandidates",(char*)&(CT->methCandidates),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethColour",(char*)&(CT->methColour),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethRelaxTSP1",(char*)&(CT->methRelaxTSP1),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethRelaxTSP2",(char*)&(CT->methRelaxTSP2),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethMaxCut",(char*)&(CT->methMaxCut),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinMethSolve",(char*)&(CT->methSolve),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethLocal",(char*)&(CT->methLocal),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMaxBBIterations",(char*)&(CT->maxBBIterations),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMaxBBNodes",(char*)&(CT->maxBBNodes),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinMethLP",(char*)&(CT->methLP),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethLPPricing",(char*)&(CT->methLPPricing),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethLPQTest",(char*)&(CT->methLPQTest),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinMethLPStart",(char*)&(CT->methLPStart),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinRandMin",(char*)&(CT->randMin),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinRandMax",(char*)&(CT->randMax),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinRandUCap",(char*)&(CT->randUCap),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinRandLCap",(char*)&(CT->randLCap),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinRandLength",(char*)&(CT->randLength),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinRandParallels",(char*)&(CT->randParallels),TCL_LINK_INT);
    Tcl_LinkVar(interp,"goblinRandGeometry",(char*)&(CT->randGeometry),TCL_LINK_INT);
    
    Tcl_LinkVar(interp,"goblinExtension",(char*)&(CT->objectExtension),
                TCL_LINK_STRING|TCL_LINK_READ_ONLY);
    
    Tcl_CreateCommand(interp,"exit",Goblin_Exit,(ClientData)NULL,
                      (Tcl_CmdDeleteProc *)NULL);

    char version[10];
    sprintf(version,"%d.%d",CT->majorVersion,CT->minorVersion);
    Tcl_PkgProvide(interp,"goblin",version);

    return TCL_OK;
}

    
// Propagate event in a thread to Tcl scripts by value changes of the
// variable goblinMasterEvent

static bool eventToggle = false;

void Goblin_MasterEventHandler()
{
    if (!eventToggle)
        Tcl_SetVar(masterInterp,"goblinMasterEvent","0",0);
    else
        Tcl_SetVar(masterInterp,"goblinMasterEvent","1",0);

    eventToggle = !eventToggle;
}


void Goblin_TraceEventHandler(char* fileName)
{
    if (CT->commLevel<3 || CT->logLevel>CT->breakLevel) return;

    MSG -> TraceAppend(fileName);
    Goblin_MasterEventHandler();
    MSG -> TraceSemTake();
}


void Goblin_LogEventHandler(msgType msg,TModule mod,THandle OH,char* logText)
{
    CT -> DefaultLogEventHandler(msg,mod,OH,logText);
    MSG -> MsgAppend(msg,mod,OH,logText);
    Goblin_MasterEventHandler();
}


bool Goblin_StopSignalHandler()
{
    return (MSG->SolverRunning() || CT->commLevel==0);
}


// Entry point for threaded script evaluation

void* Goblin_Thread(void* scriptName)
{
    Tcl_Interp* interp2 = Tcl_CreateInterp();
    Goblin_Init(interp2);
    
    MSG -> SolverSignalStarted();
    sprintf(CT->logBuffer,"Evaluating \"%s\"...",(char*)scriptName);
    CT -> LogEntry(LOG_SHELL,NoHandle,CT->logBuffer);

    #if defined(_TIMERS_)

    CT -> globalTimer[TimerSolve] -> Enable();
    
    #endif

    if (Tcl_EvalFile(interp2,(char*)scriptName)==TCL_ERROR)
    {
        sprintf(CT->logBuffer,"...Thread exit status: %s",interp2->result);
        CT -> LogEntry(LOG_IO,NoHandle,CT->logBuffer);
    }

    #if defined(_TIMERS_)

    CT -> globalTimer[TimerSolve] -> Disable();
    
    #endif

    delete[] (char*)scriptName;
    MSG -> SolverSignalIdle();
    Tcl_DeleteInterp(interp2);

    #if defined(_TIMERS_)

    sprintf(CT->logBuffer,"Thread execution time: %.0f ms",
        CT->globalTimer[TimerSolve]->PrevTime());
    CT -> LogEntry(LOG_SHELL,NoHandle,CT->logBuffer);

    #endif

    CT -> LogEntry(LOG_SHELL,NoHandle,"");

    pthread_exit(NULL);
    
    return NULL;
}


// Tcl command evaluations start here
    
int Goblin_Cmd (ClientData clientData,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[])
{
    if (argc<2)
    {
        interp->result = "Wrong number of arguments";
        return TCL_ERROR;
    }
    
    try
    {
        if (strcmp(argv[1],"mixed")==0)
        {
            if (strcmp(argv[2],"graph")!=0)
            {
                sprintf(interp->result,"Unknown option: goblin mixed %s",argv[2]);
                return TCL_ERROR;
            }

            if (argc<4 || argc>5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
            
            TNode n = 0;

            if (argc==5) n = TNode(atol(argv[4]));
       
            mixedGraph *G = new mixedGraph(n,(TOption)0,*CT);
    
            Tcl_CreateCommand(interp,argv[3],Goblin_Mixed_Graph_Cmd,(ClientData)G,
                (Tcl_CmdDeleteProc *)Goblin_Delete_Mixed_Graph);
    
            interp->result = "";
            return TCL_OK;
        }
       
        if (strcmp(argv[1],"sparse")==0)
        {
            if (argc<4 || argc>6)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
            
            TNode n = 0;

            if (argc>=5) n = TNode(atol(argv[4]));

            TNode n2 = 0;

            if (argc==6) n2 = TNode(atol(argv[5]));
       
            if (strcmp(argv[2],"graph")==0)
            {
                graph *G = new graph(n,(TOption)0,*CT);
    
                Tcl_CreateCommand(interp,argv[3],Goblin_Sparse_Graph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);
    
                interp->result = "";
                return TCL_OK;
            }
       
            if (strcmp(argv[2],"bigraph")==0)
            {
                biGraph *G = new biGraph(n,n2,(TOption)0,*CT);
    
                Tcl_CreateCommand(interp,argv[3],Goblin_Sparse_Bigraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Bigraph);
    
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"digraph")==0)
            {
                diGraph *G = new diGraph(n,(TOption)0,*CT);
    
                Tcl_CreateCommand(interp,argv[3],Goblin_Sparse_Digraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Digraph);
    
                interp->result = "";
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin sparse %s",argv[2]);
            return TCL_ERROR;
        }

        if (strcmp(argv[1],"dense")==0)
        {
            if (argc<4 || argc>6)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
            
            TNode n = 0;

            if (argc>=5) n = TNode(atol(argv[4]));

            TNode n2 = 0;

            if (argc==6) n2 = TNode(atol(argv[5]));
       
            if (strcmp(argv[2],"graph")==0)
            {
                denseGraph* G = new denseGraph(n,(TOption)0,*CT);
    
                Tcl_CreateCommand(interp,argv[3],Goblin_Dense_Graph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Graph);
    
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"bigraph")==0)
            {
                denseBiGraph* G = new denseBiGraph(n,n2,(TOption)0,*CT);
    
                Tcl_CreateCommand(interp,argv[3],Goblin_Dense_Bigraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Bigraph);
    
                interp->result = "";
                return TCL_OK;
            }
       
            if (strcmp(argv[2],"digraph")==0)
            {
                denseDiGraph* G = new denseDiGraph(n,(TOption)0,*CT);
    
                Tcl_CreateCommand(interp,argv[3],Goblin_Dense_Digraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Digraph);
    
                interp->result = "";
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin dense %s",argv[2]);
            return TCL_ERROR;
        }

        if (strcmp(argv[1],"ilp")==0 ||
            strcmp(argv[1],"lp")==0 ||
            strcmp(argv[1],"mip")==0)
        {
            if (goblinController::pLPModule==NULL)
            {
                interp->result = "No LP solver loaded";
                return TCL_ERROR;
            }

            if (argc!=5 && argc!=6)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
            
            TIndex k = TIndex(atol(argv[3]));
            TIndex l = TIndex(atol(argv[4]));
            TIndex nz = k*l/10+1;

            if (argc==6) nz = TIndex(atol(argv[5]));
       
            goblinILPModule* ModLP =
                (goblinILPModule*)goblinController::pLPModule;
            goblinILPWrapper* XLP  =
                ModLP->NewInstance(k,l,nz,goblinILPWrapper::MINIMIZE,*CT);
    
            Tcl_CreateCommand(interp,argv[2],Goblin_Ilp_Cmd,(ClientData)XLP,
                (Tcl_CmdDeleteProc *)Goblin_Delete_Ilp);
    
            interp->result = "";
            return TCL_OK;
        }
       
        if (strcmp(argv[1],"read")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
       
            Goblin_Config();
    
            goblinImport F(argv[3],*CT);
            char* type = F.Scan();
            F.DontComplain();
            
            if (strcmp(type,"mixed")==0)
            {
                mixedGraph* G = new mixedGraph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Mixed_Graph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Mixed_Graph);

                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(type,"graph")==0)
            {
                graph* G = new graph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Graph);

                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(type,"dense_graph")==0)
            {
                denseGraph* G = new denseGraph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Graph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Graph);

                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(type,"bigraph")==0)
            {
                biGraph* G = new biGraph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Bigraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Bigraph);

                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(type,"dense_bigraph")==0)
            {
                denseBiGraph* G = new denseBiGraph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Bigraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Bigraph);

                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(type,"digraph")==0)
            {
                diGraph* G = new diGraph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Digraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Sparse_Digraph);

                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(type,"dense_digraph")==0)
            {
                denseDiGraph* G = new denseDiGraph(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Digraph_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Dense_Digraph);

                interp->result = "";
                return TCL_OK;
            }
                
            if (strcmp(type,"balanced_fnw")==0)
            {
                balancedFNW* G = new balancedFNW(argv[3],*CT);

                Tcl_CreateCommand(interp,argv[2],Goblin_Balanced_FNW_Cmd,(ClientData)G,
                  (Tcl_CmdDeleteProc *)Goblin_Delete_Balanced_FNW);

                interp->result = "";
                return TCL_OK;
            }
                
            if (strcmp(type,"mixed_integer")==0)
            {
                if (goblinController::pLPModule==NULL)
                {
                    interp->result = "No LP solver loaded";
                    return TCL_ERROR;
                }

                goblinILPModule* ModLP = 
                    (goblinILPModule*)goblinController::pLPModule;
                goblinILPWrapper* XLP  =
                    ModLP->ReadInstance(argv[3],*CT);
    
                Tcl_CreateCommand(interp,argv[2],Goblin_Ilp_Cmd,(ClientData)XLP,
                    (Tcl_CmdDeleteProc *)Goblin_Delete_Ilp);

                interp->result = "";
                return TCL_OK;
            }

            sprintf(interp->result,"Not a GOBLIN object: %s",argv[3]);
            return TCL_ERROR;
        }

        if (strcmp(argv[1],"alias")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
      
            goblinAbstractObject *X = CT->ObjectPointer(THandle(atol(argv[3])));
            
            if (X->IsMixedILP())
            {
                goblinILPModule* ModLP =
                    (goblinILPModule*)goblinController::pLPModule;

                goblinILPWrapper *tmpPtr = ModLP->Reinterprete(X->BackCast());
            
                Tcl_CreateCommand(interp,argv[2],Goblin_Ilp_Cmd,tmpPtr,NULL);
                  
                interp->result = "";
                return TCL_OK;
            }

            if (!X->IsGraphObject())
            {
                sprintf(interp->result,"Not a graph object ID: %s",argv[3]);
                return TCL_ERROR;
            }

            if (X->ClassID()==IDMixedGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Mixed_Graph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            if (X->ClassID()==IDSparseGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Graph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            if (X->ClassID()==IDDenseGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Graph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            if (X->ClassID()==IDSparseBiGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Bigraph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            if (X->ClassID()==IDDenseBiGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Bigraph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            if (X->ClassID()==IDSparseDiGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Sparse_Digraph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            if (X->ClassID()==IDDenseDiGraph)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Dense_Digraph_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }
                
            if (X->ClassID()==IDBalancedFNW)
            {
                Tcl_CreateCommand(interp,argv[2],Goblin_Balanced_FNW_Cmd,
                  X->BackCast(),NULL);

                interp->result = "";
                return TCL_OK;
            }

            interp->result = "Unknown object type";
            return TCL_ERROR;
        }
    
        if (strcmp(argv[1],"export")==0)
        {
            if (strcmp(argv[2],"settings")==0)
            {
                char* configName = new char[strlen(getenv("HOME"))+9];
                sprintf(configName,"%s/.goshrc",getenv("HOME"));

                try
                {   
                    goblinExport* F = new goblinExport(configName,*CT);
                    F -> WriteConfiguration(*CT,goblinExport::CONF_FULL);
                    delete F;
                    sprintf(CT->logBuffer,
                         "...configuration file \"%s\" written",configName);
                    CT -> LogEntry(LOG_IO,NoHandle,CT->logBuffer);
                }
                catch (...) {}

                delete[] configName;
                
                return TCL_OK;
            }

            if (argc<5)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
    
            goblinController newContext(goblinDefaultContext);
            newContext.SuppressLogging();
            goblinDataObject *traceObject =
                Goblin_Read_Object(argv[3],newContext);
            newContext.traceLevel = 0;
            newContext.RestoreLogging();
    
            if (traceObject==NULL)
            {
                interp->result = "File import failed";
                return TCL_ERROR;
            }

            if (strcmp(argv[2],"ascii")==0)
            {
                TOption opt = 0;

                if (argc>5) opt = atol(argv[5]);

                traceObject -> Export2Ascii(argv[4],opt);
                delete traceObject; 
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"goblet")==0 || strcmp(argv[2],"tk")==0)
            {
                traceObject -> Export2Tk(argv[4]);
                delete traceObject; 
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"xfig")==0)
            {
                traceObject -> Export2XFig(argv[4]);
                delete traceObject; 
                interp->result = "";
                return TCL_OK;
            }

            delete traceObject;
            sprintf(interp->result,"Unknown export format: %s",argv[2]);
            return TCL_ERROR;
        }
        
        if (argc>5)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
        
        if (strcmp(argv[1],"echo")==0)
        {
            static THandle LH = NoHandle;
        
            if (strcmp(argv[2],"-nonewline")==0)
            {
                if (LH!=NoHandle) CT->LogAppend(LH,(char*)argv[3]);
                else LH = CT->LogStart(MSG_ECHO,NoHandle,(char*)argv[3]);
            }
            else 
            {
                if (LH!=NoHandle)
                {
                    CT->LogEnd(LH,(char*)argv[2]);
                    LH = NoHandle;
                }
                else CT->LogEntry(MSG_ECHO,NoHandle,(char*)argv[2]);
            }

            interp->result = "";
            return TCL_OK;
        }
    
        if (strcmp(argv[1],"solver")==0)
        {
            if (strcmp(argv[2],"thread")==0)
            {
                if (argc!=4)
                {
                    interp->result = "Wrong number of arguments";
                    return TCL_ERROR;
                }

                MSG -> SolverSignalPending();
                char* scriptName = new char[strlen(argv[3])+1];
                strcpy(scriptName,argv[3]);
                pthread_t tid[1];
                pthread_create(&tid[0],NULL,Goblin_Thread,(void*)scriptName);
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"return")==0)
            {
                if (argc!=4)
                {
                    interp->result = "Wrong number of arguments";
                    return TCL_ERROR;
                }

                while (MSG->SolverIdle()) {};

                sprintf(returnCode,"%s",argv[3]);
                errorCode = TCL_OK;
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"throw")==0)
            {
                if (argc!=4)
                {
                    interp->result = "Wrong number of arguments";
                    return TCL_ERROR;
                }
            
                while (MSG->SolverIdle()) {};

                sprintf(returnCode,"%s",argv[3]);
                errorCode = TCL_ERROR;
                interp->result = "";
                return TCL_OK;
            }

            if (argc!=3)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
        
            if (strcmp(argv[2],"stop")==0)
            {
                MSG -> SolverSignalStop();
                interp->result = "";
                return TCL_OK;
            }
        
            if (strcmp(argv[2],"idle")==0)
            {
                if (MSG->SolverIdle())
                    interp->result = "1";
                else interp->result = "0";
                
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"result")==0)
            {
                Tcl_SetResult(interp,returnCode,TCL_VOLATILE);
                return errorCode;
            }
        
            sprintf(interp->result,"Unknown option: goblin solver %s",argv[2]);
            return TCL_ERROR;
        }
    
        if (strcmp(argv[1],"messenger")==0)
        {
            if (argc!=3)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
        
            if (strcmp(argv[2],"restart")==0)
            {
                MSG -> Restart();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"reset")==0)
            {
                MSG -> MsgReset();
                interp->result = "";
                return TCL_OK;
            }

            if (strcmp(argv[2],"eof")==0)
            {
                if (MSG->MsgEOF())
                    interp->result = "1";
                else interp->result = "0";

                return TCL_OK;
            }
    
            if (strcmp(argv[2],"void")==0)
            {
                if (MSG->MsgVoid())
                    interp->result = "1";
                else interp->result = "0";

                return TCL_OK;
            }
    
            if (strcmp(argv[2],"skip")==0)
            {
                try
                {
                    MSG -> MsgSkip();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }
                
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"text")==0)
            {
                char* ret = NULL;

                try
                {
                    ret = MSG->MsgText();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }

                Tcl_SetResult(interp,ret,TCL_VOLATILE);
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"class")==0)
            {
                msgType ret = MSG_APPEND;

                try
                {
                    ret = MSG->MsgClass();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }

                sprintf(interp->result,"%d",ret);
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"module")==0)
            {
                TModule ret = NoModule;

                try
                {
                    ret = MSG->MsgModule();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }

                sprintf(interp->result,"%ld",(unsigned long)ret);
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"handle")==0)
            {
                THandle ret = NoHandle;

                try
                {
                    ret = MSG->MsgHandle();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }

                sprintf(interp->result,"%ld",(unsigned long)ret);
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"level")==0)
            {
                int ret = 0;

                try
                {
                    ret = MSG->MsgLevel();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }

                sprintf(interp->result,"%d",ret);
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"hidden")==0)
            {
                bool ret = false;

                try
                {
                    ret = MSG->MsgHidden();
                }
                catch (ERRejected)
                {
                    interp->result = "Message queue is empty";
                    return TCL_ERROR;
                }

                sprintf(interp->result,"%d",ret);
                return TCL_OK;
            }

            if (strcmp(argv[2],"filename")==0)
            {
                if (argc!=3)
                {
                    interp->result = "Wrong number of arguments";
                    return TCL_ERROR;
                }

                char* ret = MSG->TraceFilename();

                if (ret==NULL)
                {
                    interp->result = "No more queued trace files";
                    return TCL_ERROR;
                }
            
                Tcl_SetResult(interp,ret,TCL_VOLATILE);
                return TCL_OK;
            }

            if (strcmp(argv[2],"blocked")==0)
            {
                if (MSG->TraceEvent())
                    interp->result = "1";
                else interp->result = "0";
                
                return TCL_OK;
            }
    
            if (strcmp(argv[2],"unblock")==0)
            {
                MSG -> TraceUnblock();
                interp->result = "";
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin messenger %s",argv[2]);
            return TCL_ERROR;
        }
        
        if (strcmp(argv[1],"configure")==0)
        {
            CT->Configure(argc,(const char**)argv);
            interp->result = "";
            return TCL_OK;
        }
        
        #if defined(_TIMERS_)

        if (strcmp(argv[1],"timer")==0)
        {
            if (argc<3)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            if (strcmp(argv[2],"reset")==0)
            {
                CT -> ResetTimers();
                interp->result = "";
                return TCL_OK;
            }
    
            unsigned i = atol(argv[2]);

            if (i>=NoTimer)
            {
                sprintf(interp->result,"No such timer: %d",i);
                return TCL_ERROR;
            }

            if (strcmp(argv[3],"child")==0)
            {
                if (argc!=5)
                {
                    interp->result = "Wrong number of arguments";
                    return TCL_ERROR;
                }

                unsigned j = atol(argv[4]);

                if (j>=NoTimer)
                {
                    sprintf(interp->result,"No such Timer: %d",j);
                    return TCL_ERROR;
                }

                sprintf(interp->result,"%g",
                    CT->globalTimer[i]->ChildTime(TTimer(j)));
                return TCL_OK;
            }
        
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            if (strcmp(argv[3],"reset")==0)
            {
                CT -> globalTimer[i] -> Reset();
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"enable")==0)
            {
                CT -> globalTimer[i] -> Enable();
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"disable")==0)
            {
                CT -> globalTimer[i] -> Disable();
                interp->result = "";
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"label")==0)
            {
                sprintf(interp->result,"%s",
                    listOfTimers[i].timerName);
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"acc")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)CT->globalTimer[i]->AccTime());
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"av")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)CT->globalTimer[i]->AvTime());
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"max")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)CT->globalTimer[i]->MaxTime());
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"min")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)CT->globalTimer[i]->MinTime());
                return TCL_OK;
            }
    
            if (strcmp(argv[3],"prev")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)CT->globalTimer[i]->PrevTime());
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin timer <index> %s",argv[3]);
            return TCL_ERROR;
        }
        
        #endif
        
        if (strcmp(argv[1],"module")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            unsigned i = atol(argv[2]);

            if (i>=NoModule)
            {
                sprintf(interp->result,"No such module: %d",i);
                return TCL_ERROR;
            }

            if (strcmp(argv[3],"name")==0)
            {
                sprintf(interp->result,"%s",listOfModules[i].moduleName);
                return TCL_OK;
            }

            if (strcmp(argv[3],"timer")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfModules[i].moduleTimer);
                return TCL_OK;
            }

            if (strcmp(argv[3],"implementor1")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfModules[i].implementor1);
                return TCL_OK;
            }

            if (strcmp(argv[3],"implementor2")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfModules[i].implementor2);
                return TCL_OK;
            }

            if (strcmp(argv[3],"encoding_date")==0)
            {
                sprintf(interp->result,"%s",listOfModules[i].encodingDate);
                return TCL_OK;
            }

            if (strcmp(argv[3],"revision_date")==0)
            {
                sprintf(interp->result,"%s",listOfModules[i].revisionDate);
                return TCL_OK;
            }

            if (strcmp(argv[3],"original_reference")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfModules[i].originalReference);
                return TCL_OK;
            }

            if (strcmp(argv[3],"authors_reference")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfModules[i].authorsReference);
                return TCL_OK;
            }

            if (strcmp(argv[3],"text_book")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfModules[i].textBook);
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin module <index> %s",argv[3]);
            return TCL_ERROR;
        }
        
        if (strcmp(argv[1],"author")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            unsigned i = atol(argv[2]);

            if (i>=NoAuthor)
            {
                sprintf(interp->result,"No such author: %d",i);
                return TCL_ERROR;
            }

            if (strcmp(argv[3],"name")==0)
            {
                sprintf(interp->result,"%s",listOfAuthors[i].name);
                return TCL_OK;
            }

            if (strcmp(argv[3],"affiliation")==0)
            {
                sprintf(interp->result,"%s",listOfAuthors[i].affiliation);
                return TCL_OK;
            }

            if (strcmp(argv[3],"e_mail")==0)
            {
                sprintf(interp->result,"%s",listOfAuthors[i].e_mail);
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin author <index> %s",argv[3]);
            return TCL_ERROR;
        }
        
        if (strcmp(argv[1],"reference")==0)
        {
            if (argc!=4)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }

            unsigned i = atol(argv[2]);

            if (i>=NoReference)
            {
                sprintf(interp->result,"No such reference: %d",i);
                return TCL_ERROR;
            }

            if (strcmp(argv[3],"key")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].refKey);
                return TCL_OK;
            }

            if (strcmp(argv[3],"authors")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].authors);
                return TCL_OK;
            }

            if (strcmp(argv[3],"title")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].title);
                return TCL_OK;
            }

            if (strcmp(argv[3],"type")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].type);
                return TCL_OK;
            }

            if (strcmp(argv[3],"collection")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].collection);
                return TCL_OK;
            }

            if (strcmp(argv[3],"editors")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].editors);
                return TCL_OK;
            }

            if (strcmp(argv[3],"volume")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfReferences[i].volume);
                return TCL_OK;
            }

            if (strcmp(argv[3],"publisher")==0)
            {
                sprintf(interp->result,"%s",listOfReferences[i].publisher);
                return TCL_OK;
            }

            if (strcmp(argv[3],"year")==0)
            {
                sprintf(interp->result,"%ld",
                    (unsigned long)listOfReferences[i].year);
                return TCL_OK;
            }
    
            sprintf(interp->result,"Unknown option: goblin reference <index> %s",argv[3]);
            return TCL_ERROR;
        }
        
        if (strcmp(argv[1],"colour")==0)
        {
            if (argc!=3)
            {
                interp->result = "Wrong number of arguments";
                return TCL_ERROR;
            }
        
            if (strcmp(argv[2],"predecessors")==0)
            {
                sprintf(interp->result,"#%06lX",
                    CT->Colour(goblinController::PRED_COLOUR));
                return TCL_OK;
            }
        
            if (strcmp(argv[2],"infinity")==0)
            {
                sprintf(interp->result,"#%06lX",
                    CT->Colour(goblinController::INFTY_COLOUR));
                return TCL_OK;
            }

            if (strcmp(argv[2],"*")==0)
            {
                sprintf(interp->result,"#%06lX",
                    CT->Colour(goblinController::NO_COLOUR));
                return TCL_OK;
            }

            TNode c0 = atol(argv[2])+goblinController::NO_COLOUR+1;
            sprintf(interp->result,"#%06lX",CT->Colour(c0));
            return TCL_OK;
        }
    
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[1],"size")==0)
        {
            sprintf(interp->result,"%ld",goblinHeapSize);
            return TCL_OK;
        }

        if (strcmp(argv[1],"maxsize")==0)
        {
            sprintf(interp->result,"%ld",goblinMaxSize);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#allocs")==0)
        {
            sprintf(interp->result,"%ld",goblinNAllocs);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#objects")==0)
        {
            sprintf(interp->result,"%ld",goblinNObjects);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#fragments")==0)
        {
            sprintf(interp->result,"%ld",goblinNFragments);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#timers")==0)
        {
            sprintf(interp->result,"%d",NoTimer);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#authors")==0)
        {
            sprintf(interp->result,"%d",NoAuthor);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#modules")==0)
        {
            sprintf(interp->result,"%d",NoModule);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"#references")==0)
        {
            sprintf(interp->result,"%d",NoReference);
            return TCL_OK;
        }
        
        if (strcmp(argv[1],"restart")==0)
        {
            CT -> ResetCounters();
            CT -> ResetTimers();
            delete CT->logStream;
            CT->logStream = new ofstream(transscriptName);
            MSG -> Restart();

            interp->result = "";
            return TCL_OK;
        }
    }
    catch (...)
    {
        return Goblin_Propagate_Exception(interp);
    }
    
    sprintf(interp->result,"Unknown option: goblin %s",argv[1]);
    return TCL_ERROR;
}


int Goblin_Delete (ClientData clientData) throw()
{
    goblinInterprCounter--;
    if (goblinInterprCounter==0)
    {
        CT->logEventHandler = NULL;
        CT -> logStream -> flush();
        delete CT->logStream;
        CT->logStream = &clog;
        MSG -> Restart();
        delete MSG;
        delete CT;
        delete[] transscriptName;

    }
    else CT -> LogEntry(LOG_SHELL,NoHandle,"Halting GOSH interpreter...");

    return TCL_OK;
}


int Goblin_Generic_Cmd (goblinDataObject *X,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw(ERRejected,ERRange)
{
    if (argc<2)
    {
        interp->result = "Missing arguments";
        return TCL_ERROR;
    }
    
    if (strcmp(argv[1],"master")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        CT -> SetMaster(X->Handle());
        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"handle")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        sprintf(interp->result,"%ld",X->Handle());
        return TCL_OK;
    }

    if (strcmp(argv[1],"trace")==0)
    {
        if (argc!=2)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }
    
        X -> Display();
        interp->result = "";
        return TCL_OK;
    }

    if (strcmp(argv[1],"is")==0)
    {

        if (strcmp(argv[2],"mip")==0        ||
            strcmp(argv[2],"graph")==0      ||
            strcmp(argv[2],"sparse")==0     ||
            strcmp(argv[2],"directed")==0   ||
            strcmp(argv[2],"undirected")==0 ||
            strcmp(argv[2],"bipartite")==0  ||
            strcmp(argv[2],"balanced")==0
           )
        {
            interp->result = "0";
            return TCL_OK;
        }

        sprintf(interp->result,"Unknown option: %s is %s",argv[0],argv[2]);
        return TCL_ERROR;
    }

    if (strcmp(argv[1],"set")==0)
    {
        if (argc!=4)
        {
            interp->result = "Wrong number of arguments";
            return TCL_ERROR;
        }

        if (strcmp(argv[2],"label")==0)
        {
            X -> SetLabel((char*)argv[3]);
            interp->result = "";
            return TCL_OK;
        }

        sprintf(interp->result,"Unknown option: %s set %s",argv[0],argv[2]);
        return TCL_ERROR;
    }

    sprintf(interp->result,"Unknown option: %s",argv[1]);
    return TCL_ERROR;
}


int Goblin_Propagate_Exception (Tcl_Interp* interp) throw()
{
    if (CT->savedErrorMsgType != NO_MSG)
    {
        sprintf(interp->result,"%s - %s",
            CT->savedErrorMethodName,CT->savedErrorDescription);
    }
    else
    {
        CT->Error(MSG_WARN,NoHandle,"Goblin_Propagate_Exception",
            "An unknown exception has occured");
        sprintf(interp->result,"An unknown exception has occured");
    }

    CT -> logLevel = 0;
    return TCL_ERROR;
}
    

int Goblin_Exit (ClientData clientData,Tcl_Interp* interp,int argc,
    _CONST_QUAL_ char* argv[]) throw()
{
    Tcl_DeleteCommand(interp,"goblin");
            
    exit(0);
    
    return TCL_OK;
}
    

goblinDataObject *Goblin_Read_Object(const char* fileName,
    goblinController &thisContext) throw()
{
    try
    {
        goblinImport F(fileName,thisContext);
        F.DontComplain();
        
        char* type = F.Scan();
            
        if (strcmp(type,"mixed")==0)
            return new mixedGraph(fileName,thisContext);
        if (strcmp(type,"graph")==0)
            return new graph(fileName,thisContext);
        if (strcmp(type,"dense_graph")==0)
            return new denseGraph(fileName,thisContext);
        if (strcmp(type,"bigraph")==0)
            return new biGraph(fileName,thisContext);
        if (strcmp(type,"dense_bigraph")==0)
            return new denseBiGraph(fileName,thisContext);
        if (strcmp(type,"digraph")==0)
            return new diGraph(fileName,thisContext);
        if (strcmp(type,"dense_digraph")==0)
            return new denseDiGraph(fileName,thisContext);
        if (strcmp(type,"balanced_fnw")==0)
            return new balancedFNW(fileName,thisContext);
        if (strcmp(type,"mixed_integer")==0)
        {
            goblinILPModule* ModLP =
                (goblinILPModule*)goblinController::pLPModule;
            return ModLP->ReadInstance(fileName,thisContext);
        }
        
        return NULL;
    }
    catch (...) {}

    return NULL;
}
